"""
Module to manage devices
"""

from datetime import datetime
from typing import Optional, Union

import pandas as pd

from preheat_open.building_unit import BaseBuildingUnit

from .data import load_box_data
from .helpers import now, sanitise_datetime_input, timedelta


class Device(BaseBuildingUnit):
    """
    A device is a grouping of signals originating from a single physical data source (device),
    which is not linked to the building model
    """

    def __init__(self, unit_type: str, unit_data: dict, building_ref):
        unit_data_reshaped = {
            "coversBuilding": True,
            "zoneIds": [],
            "shared": False,
            "id": unit_data["id"],
            "name": unit_data["name"],
        }
        unit_data_reshaped |= {c["name"]: c for c in unit_data["components"]}

        super().__init__(
            unit_type, unit_data_reshaped, building_ref, load_data_by="cid"
        )

    def describe(
        self, display: bool = True, prefix: str = "", components: bool = True, **kwargs
    ) -> str:
        """
        Generate a text-based description of a device

        :param display: if True, prints the output
        :param prefix: string prefix to each output line (use for tabs, etc)
        :param components: if True, outputs components details
        :param kwargs: /
        :return: text description of units as string
        """
        return super().describe(
            display=display,
            prefix=prefix,
            components=components,
            show_subtype=False,
            **kwargs,
        )

    def load_state(
        self,
        seconds_back: int = 300,
        t_now=None,
        resolution_overwrite: Optional[str] = None,
    ) -> None:
        """
        Loads the state of the unit at a given point in time

        :param seconds_back: number of seconds back prior to t_now to look for data for
            (if none is found in the interval, then value will be NaN for this component)
        :param t_now: now-time override
        :param resolution_overwrite: if specified, then loads the data using that specific resolution
            (by default, data is loaded in "raw" resolution if t_now is more recent than 7 days ago, else "5min")
        :return: /
        """
        t_now = now() if t_now is None else sanitise_datetime_input(t_now)
        # Check if recent enough to expect raw data
        if resolution_overwrite is not None:
            resolution = resolution_overwrite
        elif t_now > now() - timedelta(days=7):
            resolution = "raw"
        else:
            resolution = "minute"
        self._state = load_box_data(
            self.get_all_component_cids(),
            t_now - timedelta(seconds=seconds_back),
            t_now,
            resolution,
        )

    def get_state(
        self,
        update: bool = False,
        estimate: str = "last",
        seconds_back: int = 300,
        t_now: Union[datetime, None] = None,
        resolution_overwrite: Optional[str] = None,
        **kwargs,
    ) -> pd.Series:
        """
        Fetches the state for the given unit (to force refresh it, update is required)

        :param update: if True, loads the state data to update it, else just fetches the state already loaded
        :param estimate: method to use for state estimation
            ("last" for last available value, "mean" for average in period, "median" for median value in period)
        :param seconds_back: number of seconds before t_now to look for when loading data
        :param t_now: now-time override
        :param by: whether to load data by "id" or component id "cid"
        :param resolution_overwrite: if specified, then loads the data using that specific resolution
            (by default, data is loaded in "raw" resolution if t_now is more recent than 7 days ago, else "5min")
        :return: unit state in pandas Series format
        """
        return super().get_state(
            update=update,
            estimate=estimate,
            seconds_back=seconds_back,
            t_now=t_now,
            by="cid",
            resolution_overwrite=resolution_overwrite,
        )



"""
Various helper functions
"""
import os
from copy import copy
from datetime import date, datetime, timedelta, timezone, tzinfo
from enum import Enum
from functools import lru_cache
from typing import Any, Optional, Union
from warnings import warn

import dateutil.parser
from dateutil.tz import tzutc

from preheat_open import TIMEZONE

from .types import TYPE_DATETIME_INPUT

utc = timezone.utc


@lru_cache(maxsize=1)
def DISPLAY_DEVELOPER_WARNING() -> bool:
    """
    Method used for development purposes to decide whether to show warnings for development

    :return: True if developer warnings should be enabled
    """
    from preheat_open import ENV_VAR_TRACK_MISSING_DATA_FIELD, running_in_test_mode

    try:
        env_var = os.environ.get(ENV_VAR_TRACK_MISSING_DATA_FIELD)
        out = running_in_test_mode() and (env_var is not None) and (env_var == "true")
    except Exception:
        out = False
    return out


def timestep_start(step: Union[Enum, str], t: datetime) -> datetime:
    """
    Computes the start of the current timestep

    :param step: step (check conditions below)
    :param t: time for which to evaluate the step start
    :return: start time of the timestep
    """
    if isinstance(step, Enum) and isinstance(step.value, str):
        step = step.value

    if step in ["second", "1s"]:
        t_start = t.replace(microsecond=0)
    elif step == "15s":
        sec_start = int(t.second / 15) * 15
        t_start = t.replace(microsecond=0, second=sec_start)
    elif step == "30s":
        sec_start = int(t.second / 30) * 30
        t_start = t.replace(microsecond=0, second=sec_start)
    elif step in ["minute", "1min"]:
        t_start = t.replace(microsecond=0, second=0)
    elif step == "5min":
        min_start = int(t.minute / 5) * 5
        t_start = t.replace(microsecond=0, second=0, minute=min_start)
    elif step == "15min":
        min_start = int(t.minute / 15) * 15
        t_start = t.replace(microsecond=0, second=0, minute=min_start)
    elif step == "30min":
        min_start = int(t.minute / 30) * 30
        t_start = t.replace(microsecond=0, second=0, minute=min_start)
    elif step == "hour":
        t_start = t.replace(microsecond=0, second=0, minute=0)
    elif step == "day":
        t_start = t.replace(microsecond=0, second=0, minute=0, hour=0)
    elif step == "month":
        t_start = t.replace(microsecond=0, second=0, minute=0, hour=0, day=1)
    elif step == "year":
        t_start = t.replace(microsecond=0, second=0, minute=0, hour=0, day=1, month=1)
    else:
        raise Exception("Unknown step: " + step)

    return t_start


def now(
    step: Union[Enum, str, None] = None, tz: Union[tzinfo, None] = TIMEZONE
) -> datetime:
    """
    Method to determine now time (and round it to the current step start if required)

    :param step: if provided, floors the current time to the start of the step
        (if None, no manipulation of the current time is made)
    :param tz: timezone (default: local danish timezone)
    :return: datetime
    """
    t = datetime.now(tz=tz)
    if step is None:
        return t
    else:
        return timestep_start(step, t)


def __enforce_imports():
    date.today() + timedelta(days=2)


def datetime_convert(param: TYPE_DATETIME_INPUT) -> datetime:
    """
    Converts a time-related input to a valid datetime input for other methods

    :param param: input to process
    :return: datetime corresponding to the input
    """
    if isinstance(param, datetime):
        dt = param
    elif isinstance(param, str):
        dt = dateutil.parser.parse(param)
        return dt if dt.tzinfo is not None else dt.replace(tzinfo=TIMEZONE)
    else:
        raise TypeError(f"No conversion from type: {type(param)}")

    return dt if dt.tzinfo is not None else dt.astimezone(TIMEZONE)


def sanitise_datetime_input(t: TYPE_DATETIME_INPUT) -> datetime:
    if isinstance(t, str):
        out = datetime_convert(t)
    else:
        out = t
    return out.astimezone(tzutc())


def time_resolution_aliases(resolution: str) -> Optional[str]:
    """
    Converts a resolution input to the corresponding pandas alias

    :param resolution: resolution input from methods in the package
    :return: representation in pandas (str) or None if raw or not found
    """
    if resolution in ["minute", "5min"]:
        return "5T"
    elif resolution == "hour":
        return "H"
    elif resolution == "day":
        return "D"
    elif resolution == "week":
        return "7D"  # W has poor performance, as the first day of week isn't monday
    elif resolution == "month":
        return "MS"
    elif resolution == "year":
        return "YS"
    else:
        return None


def convenience_result_list_shortener(result: list[Any]) -> Union[list[Any], Any, None]:
    """
    Method used for convenience helpers to querying methods returning lists

    :param result: input list
    :return: Depends on the input list length:
        None if list is empty,
        unique list element if list has length one (i.e. returns X if input is [X], or
        input value if the list has length > 1
    """
    n_results = len(result)
    if n_results > 1:
        return result
    elif n_results == 0:
        return None
    else:
        return result[0]


def list_to_string(list2use: list, separator: str = ",") -> str:
    """
    Helper function to turn list into string, e.g. comma separated (default)

    :param list2use: input list to convert
    :param separator: separating character
    :return: string of elements separated by the separator
    """

    if isinstance(list2use, list):
        res = separator.join(map(str, list2use))
    else:
        raise TypeError("Input list2use must be a list")
    return res


def check_no_remaining_fields(d: dict, debug_helper: Optional[str]) -> None:
    """
    Helper to check for empty kwargs (to ensure adequate propagation)

    :param d: kwargs dictionary
    :param debug_helper: string message to help debugging
    :return: /
    """
    if DISPLAY_DEVELOPER_WARNING() and len(d) > 0:
        d_in = copy(d)
        [d_in.pop(key) for key, val in d.items() if val is None]

        if len(d_in) > 0:
            if debug_helper is not None:
                msg = (
                    "There are leftover fields in "
                    + debug_helper
                    + " to be implemented: "
                )
            else:
                msg = "There are leftover fields to be implemented: "

            warn(Warning(msg + str(d_in)))


def sanitise_unit_type_input(x: str) -> str:
    """
    Convenience function to fix specific unit types with unfortunate names in API

    :param x: input
    :return: API-compatible output
    """
    if x in ["localWeatherStation", "heatPump", "pv", "carCharger"]:
        y = x + "s"
    elif x in ["secondary"]:
        y = "secondaries"
    else:
        y = x
    return y



"""
Module defining the Building class, which represents buildings containing different units and their data
"""
import re
from enum import Enum
from typing import Any, Optional, Union

import networkx as nx
import pandas as pd

from . import unit_graph
from .api import AccessDeniedError, api_get
from .backwards_compatibility import load_parameter_old_naming, warn_deprecation_pending
from .building_unit import (
    BaseBuildingUnit,
    exclude_shared_units_from_list,
    populate_units,
)
from .comfort_profiles import fetch_comfort_profiles
from .data import load_model_data
from .device import Device
from .exceptions import NoComfortProfileError
from .helpers import (
    check_no_remaining_fields,
    convenience_result_list_shortener,
    sanitise_datetime_input,
    sanitise_unit_type_input,
)
from .price import SupplyPoint
from .types import TYPE_DATETIME_INPUT
from .unit import Unit
from .weather import Weather
from .zone import Zone, populate_zones, zone_matches

# When adding new unit hierarchies, these need to be added explicitly in the list
# WARNING: Order is important due to cross-dependency
_TOP_LEVEL_UNIT_TYPES = [
    "main",
    "coldWater",
    "heating",
    "hotWater",
    "cooling",
    "electricity",
    "ventilation",
    "heatPumps",
    "indoorClimate",
    "custom",
    "localWeatherStations",
    "pvs",
    "carChargers",
]


class Building(object):
    """Building defines a building in the PreHEAT sense"""

    def __init__(self, location_id: int):
        self.location_id = location_id

        # dict containing the details of the location
        self.location: dict[str, Any] = {}
        # list of zones within the building (of class PreHEAT_API.Zone)
        self.zones: list[Zone] = []
        # dict w.units within the building (of class PreHEAT_API.Unit)
        self.units = {}
        # dict w. weather information for the location (PreHEAT_API.Weather)
        self.weather = None  # type: Weather
        # Area of the building
        self.area = None
        # All characteristics get summarised in this dictionary
        # (duplicate with some attributes do to backwards-compatibility requirement)
        self.characteristics = {}

        # Supply points (for pricing)
        self.__supply_points = []

        # Load from API
        self.__populate()

        # Construct graph
        self.__unit_graph = unit_graph.generate_unit_graph(self)

        # Traverse and map sub-unit references
        for u in self.iterate_units():
            # Add sub-units as new nodes
            if hasattr(u, "_related_sub_units_refs"):
                for ref in u._related_sub_units_refs:
                    u.add_sub_unit(self.unit_graph.nodes[ref]["unit"])
            if hasattr(u, "_related_meters_refs"):
                for ref in u._related_meters_refs:
                    u._meters.append(self.unit_graph.nodes[ref]["unit"])

        # Devices
        self.__devices: Optional[list[Device]] = None

        # Comfort profiles
        self.__comfort_profile_data = None

    def iterate_units(self) -> list[BaseBuildingUnit]:
        out = [data["unit"] for node, data in self.unit_graph.nodes.data()]
        return out

    @property
    def unit_graph(self) -> nx.MultiDiGraph:
        return self.__unit_graph

    @property
    def G(self) -> nx.MultiDiGraph:
        """DEPRECATED: use the 'unit_graph' attribute instead"""
        warn_deprecation_pending(
            "The 'G' attribute is deprecated - use the 'unit_graph' attribute instead"
        )
        return self.unit_graph

    def get_unit_graph(self) -> nx.MultiDiGraph:
        """DEPRECATED: use the 'unit_graph' attribute instead"""
        warn_deprecation_pending(
            "The 'get_unit_graph' method is deprecated - use the 'unit_graph' attribute instead"
        )
        return self.unit_graph

    def load_comfort_profile_data(
        self, start: TYPE_DATETIME_INPUT, end: TYPE_DATETIME_INPUT
    ) -> None:
        """
        Loads the comfort profile data in the building object

        :param start: start of the period
        :param end: end of the period
        """
        try:
            self.__comfort_profile_data = fetch_comfort_profiles(
                self.location_id, start=start, end=end
            )
        except NoComfortProfileError:
            pass

    def clear_comfort_profile_data(self):
        """
        Clears the comfort profile data loaded in the building object
        """
        self.__comfort_profile_data = None

    @property
    def comfort_profile_data(self) -> pd.DataFrame:
        """
        Convenience accessor to the comfort profile data for the building, if only one is defined.

        :raise: NoComfortProfileError if no data is loaded/available or if there are too many comfort profiles
            for the building
        """
        return self.get_comfort_profile_data(None)

    def get_comfort_profile_data(
        self, comfort_profile_id: Optional[int] = None
    ) -> pd.DataFrame:
        """
        Accessor to the comfort profile data for a given comfort_profile_id.
        This requires prior loading via the *load_comfort_profile_data* method.

        :param comfort_profile_id: ID of the comfort profile, if None tries to access the only comfort profile in
            the building (for convenience in small buildings without zones) and fails if too many available
        :raise: NoComfortProfileError if no data is loaded or available
        :return: dictionary of comfort profiles (key=comfort profile ID, value=dataframe[setpoint, min, max, value])
        """
        if self.__comfort_profile_data is None:
            raise NoComfortProfileError("No comfort profile loaded/available")
        elif comfort_profile_id is None:
            if len(self.__comfort_profile_data) == 1:
                out = list(self.__comfort_profile_data.values())[0]
            else:
                raise ValueError(
                    "Too many comfort profiles available (the shortcut only works when only one is available)"
                )
        else:
            out = self.__comfort_profile_data.get(comfort_profile_id)
            if out is None:
                raise NoComfortProfileError(
                    "No comfort profile available for the comfort_profile_id"
                )

        return out

    def load_data(
        self,
        start: TYPE_DATETIME_INPUT,
        end: TYPE_DATETIME_INPUT,
        resolution: Union[str, Enum] = "hour",
        components: Optional[dict[str, list[str]]] = None,
        **kwargs,
    ) -> None:
        """
        Loads timeseries data in the units of a building

        :param start: start time of the data loading
        :param end: end time of the loading
        :param resolution: resolution of the data to be loaded
        :param components: mapping of the data to be loaded (None to load everything), where the key is the unit type,
            and the value is a list of the components to load (None to load them all)
        :return: /
        """

        if components is None:
            components = {}

        start, end, resolution = load_parameter_old_naming(
            start, end, resolution, **kwargs
        )

        start = sanitise_datetime_input(start)
        end = sanitise_datetime_input(end)
        if components is None:
            components = {}

        if len(components) == 0:
            # Load all data on building, if nothing is specified
            for u in self.list_all_units():
                c_i = components.get(u.unit_type)
                u.load_data(start, end, resolution, components=c_i)
        else:
            # Otherwise, just load specific components
            for i in components.keys():
                c_i = components.get(i)
                if i == "weather":
                    self.weather.load_data(start, end, resolution, components=c_i)
                else:
                    [
                        u_i.load_data(start, end, resolution, components=c_i)
                        for u_i in self.query_units(unit_type=i)
                    ]

    def list_all_units(self) -> list[Unit]:
        out = [self.weather] + self.iterate_units()
        return out

    def load_dataset(
        self,
        component_map,
        start: TYPE_DATETIME_INPUT,
        end: TYPE_DATETIME_INPUT,
        resolution: str = "hour",
        load_weather: bool = True,
    ) -> pd.DataFrame:
        """

        :param component_map:
        :param start:
        :param end:
        :param resolution:
        :param load_weather:
        :return:
        """

        df = load_model_data(component_map, start, end, resolution)

        if load_weather is True:
            self.weather.load_data(start, end, resolution)
            df_weather = self.weather.data
            df_weather.columns = "weather." + df_weather.columns
            df = pd.concat((df_weather, df), axis=1)

        return df

    def clear_data(self) -> None:
        """
        Clears all data in the units of the object

        :return: /
        """
        self.weather.clear_data()
        [u.clear_data() for u in self.iterate_units()]

    def get_all_component_ids(self) -> dict:
        """

        :return:
        """
        return {
            k: v
            for u in self.iterate_units()
            for k, v in u.get_all_component_ids(True).items()
        }

    def get_all_component_details(self) -> list[dict[str, Union[str, float, None]]]:
        """

        :return:
        """
        all_comps = []
        for u in self.iterate_units():
            all_comps += u.get_all_component_details(prefix=True)

        return all_comps

    def query_units(
        self,
        unit_type: str = None,
        name: str = None,
        unit_id: int = None,
        exclude_shared: bool = False,
    ) -> list[BaseBuildingUnit]:
        """

        :param unit_type:
        :param name:
        :param unit_id:
        :param exclude_shared: if True, excludes shared units
        :return: a list of the relevant units
        """
        # If we pass a unit ID, find specific unit
        if unit_id:
            if unit_id in self.unit_graph.nodes.keys():
                out = [self.unit_graph.nodes[unit_id]["unit"]]
            else:
                out = []

        else:
            # If we pass a unit_type, check if we do regex search or strict search
            if unit_type:
                if unit_type[0] == "?":
                    r_type = re.compile(unit_type[1:])
                    type_match = lambda t: r_type.search(t)
                else:
                    unit_type = sanitise_unit_type_input(unit_type)
                    type_match = lambda t: t == unit_type

            # If we pass name, check if we do regex search or strict search
            if name:
                # Check if we do strict search or search by regex:
                if name[0] == "?":
                    r_name = re.compile(name[1:])
                    name_match = lambda n: r_name.search(n)
                else:
                    name_match = lambda n: n == name

            try:
                # If we pass unit_type and name
                if unit_type and name:
                    result = [
                        u
                        for u in self.iterate_units()
                        if name_match(u.name) and type_match(u.unit_type)
                    ]

                elif name:
                    result = [u for u in self.iterate_units() if name_match(u.name)]

                elif unit_type:
                    result = [
                        u for u in self.iterate_units() if type_match(u.unit_type)
                    ]

                out = result

            except Exception:
                out = []

        if exclude_shared:
            out = exclude_shared_units_from_list(out)

        return out

    def qu(self, *args, **kwargs) -> Union[BaseBuildingUnit, list[BaseBuildingUnit]]:
        """
        Convenience version of the query_units method

        :param args:
        :param kwargs:
        :return:
        """
        result = self.query_units(*args, **kwargs)
        return convenience_result_list_shortener(result)

    def __populate(self) -> None:
        # Request PreHEAT API for location
        resp = api_get(f"locations/{self.location_id}", out="json")

        # Check if we have building data and if we do, extract and populate
        if "building" in resp:
            data = resp["building"]

            self.location = data.pop("location")
            self.area = data.pop("buildingArea")
            self.type = data.pop("buildingType")
            self.characteristics = {
                "area": self.area,
                "type": self.type,
                "apartments": data.pop("apartments"),
            }

            self.weather = Weather(self.location_id, data.pop("weatherForecast"))
            self.zones = populate_zones(data.pop("zones"), building=self)

            for unit_type in _TOP_LEVEL_UNIT_TYPES:
                self.units[unit_type] = populate_units(
                    unit_type, data.pop(unit_type), self
                )

            for sp_i in data.pop("supplyPoints"):
                self.__supply_points.append(SupplyPoint(sp_i))

        else:
            raise AccessDeniedError("Access denied for given building")

        # For development validation, validate that no fields are left
        check_no_remaining_fields(data, debug_helper="building_data[building]")

    def get_supply_points(self) -> list:
        """

        :return:
        """
        return self.__supply_points

    def __repr__(self) -> str:
        return f"""{type(self).__name__}({self.location_id}): {self.location["address"]} - {self.type}"""

    def get_zones(self, zone_ids) -> list[Zone]:
        """

        :param zone_ids:
        :return:
        """
        res = []
        for z_i in self.zones:
            if z_i.id in zone_ids:
                res.append(z_i)
            res += z_i.get_sub_zones(zone_ids=zone_ids)
        return res

    def get_zone(self, id: int) -> Zone:
        """

        :param id:
        :return:
        """
        zs = self.get_zones([id])
        n_zs = len(zs)
        if n_zs < 1:
            raise Exception(f"Zone not found (id={id})")
        elif n_zs > 1:
            raise Exception(f"Too many zones found for id (id={id} / {n_zs} found)")
        return zs[0]

    def describe(self, display: bool = True) -> str:
        apartments = self.characteristics.get("apartments")
        out = """ID: {id}
Address: {address}
City: {city}
Country: {country}

Type: {building_type}
Heated area: {area}
{apartments}

Units:
==============================\n""".format(
            id=self.location_id,
            address=self.location.get("address"),
            city=self.location.get("city"),
            country=self.location.get("country"),
            building_type=self.characteristics.get("type"),
            area=self.characteristics.get("area"),
            apartments=(
                "Apartments: {}".format(apartments) if apartments is not None else ""
            ),
        )

        main_was_found = False

        for i in _TOP_LEVEL_UNIT_TYPES:
            units_i = self.query_units(i)
            if len(units_i) == 0:
                continue
            elif i == "main":
                main_was_found = True
            elif main_was_found and i in ["heating", "hotWater"]:
                continue  # Skip heating and hot water, as they are already listed under the main

            out += "    [{}] \n".format(i.upper())
            for u in units_i:
                out += u.describe(display=False, prefix="\t -", children=True) + "\n"
            out += "\n"

        out += """
Zones:
==========================\n"""
        zones = self.zones
        if len(zones) == 0:
            out += "(None)\n"
        else:
            for z in zones:
                out += z.describe(prefix="- ", children=True, display=False) + "\n"

        if display:
            print(out)
        return out

    # Managing devices
    def load_devices(self) -> None:
        """
        Loads the devices structure for the building (raw sensor structure)
        """
        resp = api_get(f"locations/{self.location_id}/devices", out="json")
        devices = resp["devices"]
        self.__devices = []
        for d in devices:
            d_type = d["typeName"]
            device_d = Device(d_type, d, self)
            self.__devices.append(device_d)

    def load_device_data(
        self,
        start: TYPE_DATETIME_INPUT,
        end: TYPE_DATETIME_INPUT,
        resolution: str = "hour",
    ):
        [d.load_data(start, end, resolution) for d in self.devices]

    def clear_device_data(
        self,
    ):
        [d.clear_data() for d in self.devices]

    @property
    def devices(self) -> list[Device]:
        if self.__devices is None:
            self.load_devices()
        return self.__devices

    def describe_devices(self, display: bool = True) -> str:
        apartments = self.characteristics.get("apartments")
        out = """ID: {id}
Address: {address}
City: {city}
Country: {country}

Type: {building_type}
Heated area: {area}
{apartments}

Devices:
==============================\n""".format(
            id=self.location_id,
            address=self.location.get("address"),
            city=self.location.get("city"),
            country=self.location.get("country"),
            building_type=self.characteristics.get("type"),
            area=self.characteristics.get("area"),
            apartments=(
                "Apartments: {}".format(apartments) if apartments is not None else ""
            ),
        )

        for d in self.devices:
            out += d.describe(display=False, prefix="\t -", children=False) + "\n"
        out += "\n"

        if display:
            print(out)
        return out

    def query_devices(
        self,
        name: Optional[str] = None,
        id: Optional[int] = None,
        device_type: Optional[str] = None,
    ) -> list[Device]:
        out = []
        for i in self.devices:
            add_i = None
            if name is not None:
                add_i = i.name == name
            if id is not None:
                if add_i is None:
                    add_i = True
                add_i = add_i and (i.id == id)
            if device_type is not None:
                if add_i is None:
                    add_i = True
                add_i = add_i and (i.unit_type == device_type)
            if add_i:
                out.append(i)
        return out

    def query_zones(
        self, zone_id: Optional[int] = None, zone_type: Optional[str] = None
    ) -> list[Zone]:
        out = []
        for z in self.zones:
            if zone_matches(z, zone_id=zone_id, zone_type=zone_type):
                out += [z]
            out += z.query_zones(zone_id=zone_id, zone_type=zone_type)
        return out


def available_buildings() -> pd.DataFrame:
    """
    lists available buildings

    :return: dataframe of buildings available
    :rtype:
    """
    out = api_get("locations")
    return pd.DataFrame(out["locations"])



from typing import Optional

import pandas as pd

from .backwards_compatibility import load_parameter_old_naming
from .component import Component
from .data import load_weather_data
from .helpers import check_no_remaining_fields, now, sanitise_datetime_input, timedelta
from .types import TYPE_DATETIME_INPUT
from .unit import Unit


class Weather(Unit):
    """Weather; an extension of Unit to handle Weather extraction"""

    def __init__(self, location_id, weather_data):
        super().__init__("weather", weather_data)
        self.location_id = location_id
        self.id = weather_data.pop("gridId", None)
        self.name = "WeatherForecast"

        [
            self.components.append(Component(type_i))
            for type_i in weather_data.pop("types")
        ]
        check_no_remaining_fields(weather_data, debug_helper="weather_unit_data")

    def load_data(
        self,
        start: TYPE_DATETIME_INPUT,
        end: TYPE_DATETIME_INPUT,
        resolution: str = "hour",
        components=None,
        **kwargs
    ) -> None:
        # Parse strings to datetime object

        start, end, resolution = load_parameter_old_naming(
            start, end, resolution, **kwargs
        )

        start = sanitise_datetime_input(start)
        end = sanitise_datetime_input(end)

        self._warn_if_data_is_loaded()
        components_to_load = self.get_all_component_ids(components=components)
        if components is not None:
            components_to_load = {
                k: components_to_load[k]
                for k in components_to_load
                if components_to_load[k] in components
            }

        if (end - start) > pd.Timedelta("180d"):
            # Batch in 6M intervals (max we can load at a time)
            parts = list(pd.date_range(start=start, end=end, freq="180d"))

            if start != parts[0]:
                parts.insert(0, start)
            if end != parts[-1]:
                parts.append(end)

            parts[0] -= pd.Timedelta("1d")  # we add back one day later

            if parts[-1] - parts[-2] < pd.Timedelta("1d"):
                parts[-2] -= pd.Timedelta("1d")

            pairs = zip(map(lambda d: d + pd.Timedelta("1d"), parts[:-1]), parts[1:])

            dfs = [
                load_weather_data(
                    self.location_id,
                    components_to_load,
                    start,
                    end,
                    resolution,
                )
                for start, end in pairs
            ]
            res = pd.concat(dfs)

        else:
            res = load_weather_data(
                self.location_id,
                components_to_load,
                start,
                end,
                resolution,
            )

        self.data: pd.DataFrame = res
        self._ensure_continuity_of_data(resolution)

    def load_state(
        self,
        seconds_back: int = 3600,
        t_now=None,
        resolution_overwrite: Optional[str] = None,
    ) -> None:
        t_now = now() if t_now is None else sanitise_datetime_input(t_now)

        if resolution_overwrite is None:
            resolution = "hour"
        else:
            resolution = resolution_overwrite

        components_to_load = self.get_all_component_ids()
        self._state = load_weather_data(
            self.location_id,
            components_to_load,
            t_now - timedelta(seconds=seconds_back),
            t_now,
            resolution,
        )

    def get_state(
        self,
        update: bool = False,
        estimate: str = "last",
        seconds_back: int = 3600,
        resolution_overwrite: Optional[str] = None,
        **kwargs
    ) -> pd.Series:
        return super().get_state(
            update=update,
            estimate=estimate,
            seconds_back=seconds_back,
            resolution_overwrite=resolution_overwrite,
            **kwargs
        )



"""
To manage timeseries data loading
"""
import math
from datetime import datetime, timedelta
from functools import reduce
from typing import Any, Optional

import numpy as np
import pandas as pd

from .api import (
    DATETIME_FORMAT,
    MAX_IDS_AND_CIDS_PER_REQUEST,
    MAX_POINTS_PER_REQ,
    APIDataExtractError,
    api_get,
    datetime_to_api_string,
)
from .helpers import list_to_string, sanitise_datetime_input
from .types import TYPE_DATETIME_INPUT

__DEFAULT_DATAFRAME_INDEX = pd.DatetimeIndex([], tz="UTC")


def load_box_data(
    component_map: dict[str, Any],
    start: TYPE_DATETIME_INPUT,
    end: TYPE_DATETIME_INPUT,
    resolution="minute",
) -> pd.DataFrame:
    """
    Loads box data from the API

    :param component_map: mapping of the components to load data for
    :param start: start of the period
    :param end: end of the period
    :param resolution: resolution to load data with (raw, minute/5min, hour , day, week, month, year)
    :return: data in Dataframe format
    """
    start = sanitise_datetime_input(start)
    end = sanitise_datetime_input(end)
    __check_start_end_date(start, end)

    path = "measurements"
    cid_key = "cids"
    cids = list(component_map.keys())

    if not cids:
        return pd.DataFrame(data={}, index=__DEFAULT_DATAFRAME_INDEX)

    resp = __perform_requests(path, cid_key, cids, start, end, resolution, "csv")

    return __extract_data(resp, "cid", component_map, "csv").astype(float)


def load_model_data(
    component_map: dict[str, Any],
    start: TYPE_DATETIME_INPUT,
    end: TYPE_DATETIME_INPUT,
    resolution="raw",
) -> pd.DataFrame:
    """
    Loads data using a building model representation

    :param component_map: mapping of the components
    :param start: start of the period
    :param end: end of the period
    :param resolution: resolution of the data (raw, minute/5min, hour, day, week, month, year)
    :return: data loaded in dataFrame format
    """
    start = sanitise_datetime_input(start)
    end = sanitise_datetime_input(end)
    __check_start_end_date(start, end)

    path = "units/measurements"
    id_key = "ids"
    ids = list(component_map.keys())

    if not ids or len(ids) == 0:
        return pd.DataFrame(data={}, index=__DEFAULT_DATAFRAME_INDEX)

    if resolution != "raw":
        # Do chunked requests
        resp = __perform_requests(path, id_key, ids, start, end, resolution, "csv")
    else:
        # Do single request
        resp = __make_request(path, {id_key: ids}, start, end, resolution, "csv")

    return __extract_data(resp, "id", component_map, "csv").astype(float)


def load_weather_data(
    location_id: int,
    component_map: dict[str, Any],
    start: TYPE_DATETIME_INPUT,
    end: TYPE_DATETIME_INPUT,
    resolution: str = "hour",
    **kwargs,
) -> pd.DataFrame:
    """
    Loads weather data from the API

    :param location_id: location ID
    :param component_map: mapping of the components
    :param start: start of the period
    :param end: end of the period
    :param resolution: resolution of the data (raw, minute/5min, hour, day, week, month, year)
        Note: raw, minute/5min are overwritten to hour
    :param kwargs: /
    :return: loaded data in DataFrame format
    """
    # Parse strings to datetime object
    start = sanitise_datetime_input(start)
    end = sanitise_datetime_input(end)
    __check_start_end_date(start, end)

    path = f"weather/{location_id}"
    cid_key = "type_ids"
    cids = list(component_map.keys())

    # If minutes, then we need to resample
    if resolution in ["minute", "5min", "raw"]:
        resolution = "hour"
        t_span = (start - timedelta(hours=1), end + timedelta(hours=1), resolution)
        resample = True
    else:
        t_span = (start, end, resolution)
        resample = False

    resp = __perform_requests(path, cid_key, cids, *t_span, "csv")

    if resample:
        return (
            __extract_data(resp, "type_id", component_map, "csv")
            .resample("5min")
            .ffill()[start:end]
        )

    return __extract_data(resp, "type_id", component_map, "csv").astype(float)


def perform_requests(
    path,
    cid_key,
    cids,
    start: TYPE_DATETIME_INPUT,
    end: TYPE_DATETIME_INPUT,
    resolution: str,
    format: str = "csv",
):
    return __perform_requests(path, cid_key, cids, start, end, resolution, format)


def __perform_requests(
    path,
    cid_key,
    cids,
    start: TYPE_DATETIME_INPUT,
    end: TYPE_DATETIME_INPUT,
    resolution: str,
    format: str = "csv",
):
    # Split CIDs into chunks
    # NOTE: we reduce number of CIDs pr. REQ to confine w. MAX_POINTS_PER_REQ
    # This sets the limit to 90000 points on a single CID; which for 5min.
    # is roughly 300 days of data.
    time_lookup = {
        "raw": 10,
        "minute": 300,
        "5min": 300,
        "hour": 3600,
        "day": 86400,
        "week": 604800,
        "month": 16934400,
        "year": 6181056000,
    }

    end_date_in = sanitise_datetime_input(end)
    start_date_in = sanitise_datetime_input(start)

    dt = end_date_in - start_date_in

    N_points_requested = len(cids) * dt.total_seconds() / time_lookup[resolution]

    if N_points_requested < 1:
        reqs = []
    else:
        cid_chunk_size = (MAX_POINTS_PER_REQ / N_points_requested) * len(cids)

        # Imposing another max of number of IDs and CIDs regardless of resolution
        cid_chunk_size = min(cid_chunk_size, MAX_IDS_AND_CIDS_PER_REQUEST)

        cid_chunks = [
            cids[x : x + max(int(cid_chunk_size), 1)]
            for x in range(0, len(cids), max(int(cid_chunk_size), 1))
        ]
        time_chunks = math.ceil(1 / cid_chunk_size)
        dt_chunk = (end_date_in - start_date_in) / np.fmax(1, time_chunks)

        reqs = []
        for cid_chunk in cid_chunks:
            for i in range(time_chunks):
                reqs.append(
                    __make_request(
                        path,
                        {cid_key: list_to_string(cid_chunk)},
                        start + i * dt_chunk,
                        start + (i + 1) * dt_chunk,
                        resolution,
                        format,
                    )
                )

    return reqs


def __make_request(
    path, payload, start: datetime, end: datetime, resolution: str, format: str = "csv"
):
    payload["start_time"] = datetime_to_api_string(start)
    payload["end_time"] = datetime_to_api_string(end)

    if resolution == "5min":
        resolution = "minute"

    payload["time_resolution"] = resolution

    return api_get(path, payload=payload, out=format)


def extract_data(resp, id_key, component_map, format: str = "csv"):
    return __extract_data(resp, id_key, component_map, format)


def __reformat_dataframe(
    resp: pd.DataFrame,
    component_map: dict[str, str],
    id_key: Optional[str] = None,
) -> pd.DataFrame:
    kwargs_to_datetime = dict(utc=True, format=DATETIME_FORMAT)
    # If this is having an issue, use "infer_datetime_format=True" instead of "format=..."

    # Use of resp.loc["time"] was removed below as it triggers warnings for pandas>=1.5
    resp["time"] = pd.to_datetime(resp["time"], **kwargs_to_datetime)

    resp_columns = resp.keys()
    if id_key is not None:
        components = id_key
    elif "id" in resp_columns:
        components = "id"
    elif "type_id" in resp_columns:
        components = "type_id"
    else:
        raise RuntimeError("Unknown column formatting")
    comp_map_rename = {int(i): component_map[i] for i in component_map}
    component_names = comp_map_rename.values()
    if resp.empty:
        # For backwards compatibility
        out = pd.DataFrame(
            {i: [] for i in component_names}, index=__DEFAULT_DATAFRAME_INDEX
        )
    else:
        out = resp.pivot(index="time", columns=components, values="value").rename(
            columns=comp_map_rename,
        )

    for c in component_names:
        if c not in out.keys():
            out[c] = np.nan

    out.axes[1].name = None  # For backwards compatibility
    return out


def __extract_data(resp, id_key, component_map, format="csv"):
    if len(resp) < 1:
        # Short circuit all in case of empty response
        out = pd.DataFrame(
            {i: [] for i in component_map.values()}, index=__DEFAULT_DATAFRAME_INDEX
        )
    elif format == "json":
        if isinstance(resp, list):
            # Initialize empty response that we can append to
            resp_new = {cid: [] for cid in component_map.keys()}
            for req in resp:
                for cid, values in req.items():
                    resp_new[cid] += values
            resp = resp_new

        if "message" in resp:
            raise APIDataExtractError(resp["message"])

        kwargs_to_datetime = dict(utc=True, format=DATETIME_FORMAT)
        dfs = []
        for cid, name in component_map.items():
            time = pd.to_datetime(
                [data["time"] for data in resp[cid]], **kwargs_to_datetime
            )
            value = [data["value"] for data in resp[cid]]

            df_i = pd.DataFrame(value, columns=[name], index=time)
            dfs.append(df_i)

        out = reduce(
            lambda x, y: pd.merge(x, y, left_index=True, right_index=True, how="outer"),
            dfs,
        )

    elif format == "csv":
        # Join response into one
        if isinstance(resp, list):
            resp_new = None
            for df in resp:
                resp_new = df if resp_new is None else pd.concat([resp_new, df])
            resp = resp_new

        out = __reformat_dataframe(resp, component_map, id_key)

    elif format == "dataframe":
        out = __reformat_dataframe(resp, component_map, id_key)

    else:
        raise TypeError("Unsupported format")

    out.index.rename(None, inplace=True)
    return out


def __check_start_end_date(start: datetime, end: datetime) -> None:
    if end < start:
        raise ValueError("End date must be AFTER start date")



"""
The preheat_open package provides methods to load building data from `Neogrids <https://www.neogrid.dk/>`_ platform
"""

# Imports for backwards compatibility of dependent code
# noinspection PyUnresolvedReferences
import os
import re
import sys

from . import _version
from .api import (
    BASE_URL,
    MAX_CIDS_PER_REQ,
    MAX_POINTS_PER_REQ,
    TIMEZONE,
    AccessDeniedError,
    APIDataExtractError,
    APIKeyMissingError,
    load_configuration,
    set_api_key,
)

# noinspection PyUnresolvedReferences
from .building import Building, available_buildings

# noinspection PyUnresolvedReferences
from .logging import set_logging_level

# noinspection PyUnresolvedReferences
from .unit import Unit

__version__ = _version.get_versions()["version"]


config = load_configuration()


def running_in_test_mode() -> bool:
    """
    Method to evaluate whether the code is running in test mode. Test mode is activated when using Pytest

    :return: True if running in test mode, False otherwise
    """
    # Pytest is detailed via argv or an environment variable
    if "PYTEST_CURRENT_TEST" in os.environ:
        return True
    elif any(re.findall(r"pytest|py.test", sys.argv[0])):
        return True
    else:
        return False


# list of environment variables for tricks
ENV_VAR_TRACK_MISSING_DATA_FIELD = "PREHEAT_OPEN_TRACK_MISSING_FIELDS"



import inspect
import logging
import os
import sys
import warnings
from functools import lru_cache
from typing import Callable, Optional, Union

from .singleton import Singleton

LOG_FORMAT = "%(asctime)-23s  %(levelname)-8s  %(name)-32s  %(message)-160s  .(%(filename)s:%(lineno)d)"


class Logging(metaclass=Singleton):
    def __init__(self, stacklevel: int = 0):
        # Ensure the routing of warnings to the logger when using the logger
        logging.captureWarnings(True)

        def get_logger(name):
            return logging.getLogger().getChild(name)

        self.get_root_logger: Callable[[str], logging.Logger] = get_logger
        if sys.gettrace() is None:
            logging.basicConfig(level=logging.INFO, format=LOG_FORMAT)
        else:
            logging.basicConfig(level=logging.DEBUG, format=LOG_FORMAT)

    def __fix_get_root_logger(self, x):
        """Quick fix to support 2 structures of root logger"""
        try:
            out = self.get_root_logger(x)
        except Exception:
            out = self.get_root_logger()
        return out

    def _get_logger(self, failsafe: bool = True, stack_level: int = 0):
        try:
            frame = inspect.stack()[2 + stack_level]
            mod = inspect.getmodule(frame[0])
            # return self.get_root_logger(mod.__package__)
            return self.__fix_get_root_logger(mod.__name__.split(".")[-1])
        except Exception as e:
            if failsafe is True:
                warnings.warn(
                    "There was an issue in accessing the desired logger - defaulting to a failsafe one"
                )
                return self.__fix_get_root_logger("Failsafe-log (preheat_open)")
            else:
                raise e

    def set_root_logger(self, func: Callable[[str], logging.Logger]):
        self.get_root_logger = func

    def set_level(self, level: Union[str, int]) -> None:
        logging.basicConfig(level=level, force=True)

    def critical(
        self,
        msg: str,
        exception: Optional[Exception] = None,
        stack_level: int = 0,
        *args,
        **kwargs,
    ) -> None:
        if exception is None:
            [
                self._get_logger(stack_level=stack_level).critical(
                    line, stacklevel=2 + stack_level, *args, **kwargs
                )
                for line in str(msg).split("\n")
            ]

        else:
            try:
                raise exception
            except Exception as e:
                [
                    self._get_logger(stack_level=stack_level).exception(
                        line, stacklevel=2 + stack_level, *args, **kwargs
                    )
                    for line in str(msg).split("\n")
                ]

    def error(
        self,
        msg: str,
        exception: Optional[Exception] = None,
        stack_level: int = 0,
        *args,
        **kwargs,
    ) -> None:
        if exception is None:
            [
                self._get_logger(stack_level=stack_level).error(
                    line, stacklevel=2 + stack_level, *args, **kwargs
                )
                for line in str(msg).split("\n")
            ]

        else:
            try:
                raise exception
            except Exception as e:
                [
                    self._get_logger(stack_level=stack_level).exception(
                        line, stacklevel=2 + stack_level, *args, **kwargs
                    )
                    for line in str(msg).split("\n")
                ]

    def warning(
        self, msg: Union[str, Warning], stack_level: int = 0, *args, **kwargs
    ) -> None:
        if isinstance(msg, Warning):
            warnings.warn(msg)
        else:
            [
                self._get_logger(stack_level=stack_level).warning(
                    line, stacklevel=2 + stack_level, *args, **kwargs
                )
                for line in str(msg).split("\n")
            ]

    def info(self, msg: str, stack_level: int = 0, *args, **kwargs) -> None:
        [
            self._get_logger(stack_level=stack_level).info(
                line, stacklevel=2 + stack_level, *args, **kwargs
            )
            for line in str(msg).split("\n")
        ]

    def debug(self, msg: str, stack_level: int = 0, *args, **kwargs) -> None:
        [
            self._get_logger(stack_level=stack_level).debug(
                line, stacklevel=2 + stack_level, *args, **kwargs
            )
            for line in str(msg).split("\n")
        ]


def logging_level(level: str) -> int:
    """
    Converts a string to a logging level

    :param level: logging level (debug, info, warning, error, critical)
    :type level:
    :return: logging level identifier (in logging package)
    :rtype:
    """
    if isinstance(level, str):
        str_level = level.lower()
        if str_level == "debug":
            log_level = logging.DEBUG
        elif str_level == "info":
            log_level = logging.INFO
        elif str_level == "warning":
            log_level = logging.WARNING
        elif str_level == "error":
            log_level = logging.ERROR
        elif str_level == "critical":
            log_level = logging.CRITICAL
        else:
            raise Exception(f"Illegal logging level ({level})")

        return log_level

    else:
        raise Exception("Only logging levels in string format are supported for now")


def set_logging_level(level: str) -> None:
    """
    Sets the logging level

    :param level: logging level (debug, info, warning, error, critical)
    :type level:
    """
    Logging().set_level(logging_level(level))


@lru_cache(maxsize=1)
def __profiling_log_file_path() -> Optional[str]:
    return os.environ.get("PREHEAT_OPEN_LOG_PROFILER")


@lru_cache(maxsize=1)
def write_log_to_profiling_file_is_active() -> bool:
    """
    Method to determine whether to enable the logging to a profiling file
    (enable by setting the environment variable "PREHEAT_OPEN_LOG_PROFILER")
    """
    v = __profiling_log_file_path()
    return v is not None


def write_log_to_profiling_file(message: str) -> None:
    """
    Method to write a line to the profiling file, which can be used to track which kind of I/O is used
    (enable by setting the environment variable "PREHEAT_OPEN_LOG_PROFILER")
    """
    if not write_log_to_profiling_file_is_active():
        return

    with open(__profiling_log_file_path(), "a") as f:
        f.write(message + "\n")



import pytest

from preheat_open.backwards_compatibility import load_parameter_old_naming


@pytest.mark.parametrize(
    "input_param,expected",
    [
        (
            {"start": "2022-01-01", "end": "2022-01-02", "resolution": "hour"},
            ("2022-01-01", "2022-01-02", "hour"),
        ),
        (
            {
                "start": "2022-01-01",
                "end": "2022-01-02",
                "resolution": "hour",
                "time_resolution": "minute",
            },
            ("2022-01-01", "2022-01-02", "minute"),
        ),
        (
            {
                "start": "2022-01-01",
                "end": "2022-01-02",
                "resolution": "hour",
                "start_date": "2022-01-03",
                "end_date": "2022-01-04",
            },
            ("2022-01-03", "2022-01-04", "hour"),
        ),
    ],
)
def test_load_parameter_old_naming(input_param, expected):
    if "start_date" in input_param.keys() or "time_resolution" in input_param.keys():
        with pytest.warns(DeprecationWarning):
            assert expected == load_parameter_old_naming(**input_param)
    else:
        assert expected == load_parameter_old_naming(**input_param)



"""
Management control units and scheduling
"""
from typing import Any

import numpy as np
import pandas as pd
from requests.models import Response

from .building_unit import BaseBuildingUnit
from .helpers import check_no_remaining_fields
from .logging import Logging
from .setpoints import (
    InvalidScheduleError,
    get_setpoint_schedule,
    send_setpoint_schedule,
)
from .types import TYPE_DATETIME_INPUT


def validate_schedule(schedule_df: pd.DataFrame) -> None:
    """
    Method to validate a schedule (basically checking that there is no NaN or Inf data, and that the schedule is not empty)

    :param schedule_df: schedule to validate
    :raise: InvalidScheduleError if the schedule is invalid
    :return: /
    """
    if schedule_df["value"].isnull().any():
        raise InvalidScheduleError("requested schedule has missing values")
    if np.isinf(schedule_df["value"]).any():
        raise InvalidScheduleError("requested schedule has infinite values")
    if schedule_df.empty or (len(schedule_df) == 0):
        raise InvalidScheduleError(
            "An empty schedule_df was requested (the API does not accept this)"
        )


class ControlUnit(BaseBuildingUnit):
    """Control Unit; an extension of Unit to handle controls"""

    def __init__(self, unit_data: dict[str, Any], building_ref=None):
        super().__init__("control", unit_data, building_ref)
        if "active" in unit_data.keys():
            self.active = unit_data.pop("active")
        else:
            self.active = False

        check_no_remaining_fields(unit_data, debug_helper="control_unit_data")

    def request_schedule(self, schedule_df: pd.DataFrame) -> Response:
        """
        Sends a scheduling request to the API

        :param schedule_df: schedule to execute, as a DataFrame
        :return: requests.Response for the scheduling request
        """
        if self.active is False:
            Logging().warning(
                RuntimeWarning(
                    """Warning: you are trying to control an unit that is not activated 
                    (id={} / details: [unit: {} / building: [{}] {}])""".format(
                        self.id,
                        self.name,
                        self.building.location["locationId"],
                        self.building.location["address"],
                    )
                )
            )
        validate_schedule(schedule_df)
        return send_setpoint_schedule(self.id, schedule_df)

    def get_schedule(
        self, start_date: TYPE_DATETIME_INPUT, end_date: TYPE_DATETIME_INPUT
    ) -> pd.DataFrame:
        """
        Loads the past schedule for the unit and a given period

        :param start_date: start of the period
        :param end_date: end of the period
        :return: schedule in DataFrame format
        """
        return get_setpoint_schedule(self.id, start_date, end_date)



def test_import():
    import preheat_open
    from preheat_open import (
        api,
        backwards_compatibility,
        building,
        building_unit,
        cache,
        comfort_profiles,
        component,
        control_unit,
        data,
        device,
        exceptions,
        helpers,
        logging,
        price,
        setpoints,
        singleton,
        types,
        unit,
        unit_graph,
        weather,
        zone,
    )

    # Test top-level module
    assert isinstance(preheat_open.__version__, str)
    assert isinstance(preheat_open.__name__, str)

    # Test sub-modules
    assert isinstance(api.__name__, str)
    assert isinstance(backwards_compatibility.__name__, str)
    assert isinstance(building.__name__, str)
    assert isinstance(building_unit.__name__, str)
    assert isinstance(cache.__name__, str)
    assert isinstance(comfort_profiles.__name__, str)
    assert isinstance(component.__name__, str)
    assert isinstance(control_unit.__name__, str)
    assert isinstance(data.__name__, str)
    assert isinstance(device.__name__, str)
    assert isinstance(exceptions.__name__, str)
    assert isinstance(helpers.__name__, str)
    assert isinstance(logging.__name__, str)
    assert isinstance(price.__name__, str)
    assert isinstance(setpoints.__name__, str)
    assert isinstance(singleton.__name__, str)
    assert isinstance(types.__name__, str)
    assert isinstance(unit.__name__, str)
    assert isinstance(unit_graph.__name__, str)
    assert isinstance(weather.__name__, str)
    assert isinstance(zone.__name__, str)



from pandas import DataFrame, Series

from preheat_open import test
from preheat_open.device import Device


class TestBuildingDevice(test.PreheatTest):
    def test_class(self, building_with_devices, short_period):
        devices = building_with_devices.devices

        def _validate_list(l):
            assert isinstance(l, list) and isinstance(l[0], Device) and (len(l) == 1)

        _validate_list(building_with_devices.query_devices(name="Fake test device"))
        _validate_list(building_with_devices.query_devices(device_type="HEAT_METER"))
        _validate_list(building_with_devices.query_devices(id=13255))
        _validate_list(
            building_with_devices.query_devices(
                name="Fake test device", device_type="HEAT_METER", id=13255
            )
        )

        assert (
            len(
                building_with_devices.query_devices(
                    name="Fake", device_type="HEAT_METER", id=13255
                )
            )
            == 0
        )
        assert (
            len(
                building_with_devices.query_devices(
                    name="Fake test device", device_type="0", id=13255
                )
            )
            == 0
        )
        assert (
            len(
                building_with_devices.query_devices(
                    name="Fake test device", device_type="HEAT_METER", id=1
                )
            )
            == 0
        )

        assert len(building_with_devices.query_devices(id=1)) == 0
        assert len(building_with_devices.query_devices(name="Does not exist")) == 0
        assert (
            len(building_with_devices.query_devices(device_type="Fake test device"))
            == 0
        )

        building_with_devices.load_device_data(*short_period)
        df0 = devices[0].data
        assert isinstance(df0, DataFrame)
        assert len(df0) > 1

        building_with_devices.clear_device_data()
        assert devices[0].data.empty

        s = devices[0].get_state(update=True, t_now=short_period[1])
        assert isinstance(s, Series) and len(s) > 0



from copy import deepcopy

import pytest

import preheat_open
from preheat_open.test import (
    API_KEY,
    MEDIUM_TEST_PERIOD,
    SHORT_TEST_PERIOD,
    TEST_LOCATION_ID,
    mark_pytest_package_test,
)

# Defining fixture scope
FIXTURE_SCOPE = "session"


# Fixing API key once and for all
@pytest.fixture(autouse=True)
def mark_as_pytest_run():
    mark_pytest_package_test()


@pytest.fixture(autouse=True)
def set_api_key():
    preheat_open.api.set_api_key(API_KEY)


@pytest.fixture(scope=FIXTURE_SCOPE)
def medium_period():
    return MEDIUM_TEST_PERIOD


@pytest.fixture(scope=FIXTURE_SCOPE)
def short_period():
    return SHORT_TEST_PERIOD


@pytest.fixture(scope=FIXTURE_SCOPE)
def building():
    return preheat_open.Building(TEST_LOCATION_ID)


@pytest.fixture(scope=FIXTURE_SCOPE)
def building_with_devices(building):
    b = deepcopy(building)
    b.load_devices()
    return b


@pytest.fixture(scope=FIXTURE_SCOPE)
def building_with_data(building, medium_period):
    building_new = deepcopy(building)
    building_new.load_data(*medium_period)
    return building_new


# Legacy fixtures
@pytest.fixture(scope=FIXTURE_SCOPE)
def building_id():
    return TEST_LOCATION_ID


@pytest.fixture(scope=FIXTURE_SCOPE)
def unit_id():
    return 15312


@pytest.fixture(scope=FIXTURE_SCOPE)
def control_unit_id():
    return 15357


@pytest.fixture(scope=FIXTURE_SCOPE)
def unit(building, unit_id):
    return building.query_units(unit_id=unit_id)[0]


@pytest.fixture(scope=FIXTURE_SCOPE)
def unit_with_data(unit, medium_period):
    unit_new = deepcopy(unit)
    unit_new.load_data(*medium_period)
    return unit_new


@pytest.fixture(scope=FIXTURE_SCOPE)
def control_unit(building):
    return building.qu("control", "control_unit_custom_1")


@pytest.fixture(scope=FIXTURE_SCOPE)
def weather_unit(building):
    return building.weather



import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from preheat_open import test
from preheat_open.exceptions import NoComfortProfileError


class TestBuilding(test.PreheatTest):
    def test_zone_type(self, building):
        apt_zone = building.get_zones(zone_ids=[4832])[0]
        wet_zone = building.get_zones(zone_ids=[4833])[0]
        dry_zone = building.get_zones(zone_ids=[4835])[0]

        type_1, wet_or_dry_1 = apt_zone.get_type()
        assert type_1 == "apartment"
        assert wet_or_dry_1 == "?"

        type_a, wet_or_dry_a = wet_zone.get_type()
        assert type_a == "bathroom"
        assert wet_or_dry_a == "wet"

        type_b, wet_or_dry_b = dry_zone.get_type()
        assert type_b == "corridor"
        assert wet_or_dry_b == "dry"

    def query_zones(self, building):
        z = building.query_zones(zone_id=4832, zone_type="apartment")[0]
        assert len(z.query_zones(zone_id=4833)) == 1
        assert len(z.query_zones(zone_type="bathroom")) == 1
        assert len(z.query_zones(zone_id=4833, zone_type="bathroom")) == 1

        assert len(z.query_zones(zone_id=4833, zone_type="apartment")) == 0
        assert len(z.query_zones(zone_type="doesnt exist")) == 0
        assert len(z.query_zones(zone_id=0)) == 0

    def test_get_units(self, building_with_data):
        # Zone A1 -- connected to unit named 'indoor_climate_1'
        zone_A1 = building_with_data.get_zones(zone_ids=[4830])[0]
        units_A1 = zone_A1.get_units("indoorClimate")
        assert units_A1[0] == building_with_data.qu(name="indoor_climate_1")
        zone_A1b = building_with_data.get_zone(id=4830)
        assert zone_A1 == zone_A1b

        # Zone A2 -- not connected to any units
        zone_A2 = building_with_data.get_zones(zone_ids=[4831])[0]
        units_A2 = zone_A2.get_units("indoorClimate")
        assert not units_A2

        zone_A = zone_A1b.get_parent_zone()
        assert zone_A.get_parent_zone() is None
        assert len(zone_A.get_units("indoorClimate")) > 0

    def test_comfort_profiles(self, building):
        # Zone A1 -- connected to unit named 'indoor_climate_1'
        zone_A1 = building.get_zones(zone_ids=[4830])[0]

        start_cp = "2023-01-05"
        end_cp = "2023-01-06"
        zone_A1.load_comfort_profile_data(start_cp, end_cp)

        df_cp = zone_A1.comfort_profile_data
        assert isinstance(df_cp, pd.DataFrame)
        assert 23 <= len(df_cp) <= 25

        df_cp2 = zone_A1.sub_zones[0].comfort_profile_data
        assert isinstance(df_cp2, pd.DataFrame)
        assert_frame_equal(df_cp, df_cp2)

        with pytest.raises(NoComfortProfileError):
            zone_A1.get_parent_zone().comfort_profile_data

    def test_get_unit_types(self, building_with_data):
        # Zone A1 -- connected to unit named 'indoor_climate_1'
        zone_A1 = building_with_data.get_zones(zone_ids=[4830])[0]
        types_A1 = zone_A1.get_unit_types()
        assert isinstance(types_A1, list)
        for t in types_A1:
            assert isinstance(t, str)

        # Zone A2 -- not connected to any units
        zone_A2 = building_with_data.get_zones(zone_ids=[4831])[0]
        types_A2 = zone_A2.get_unit_types()
        assert isinstance(types_A2, list)
        assert not types_A2



from typing import Any

from .backwards_compatibility import warn_deprecation_pending


class Component(object):
    """Defines a Component in the PreHEAT sense"""

    def __init__(self, component_data: dict[str, Any]):
        # Identifier of the component
        self.cid = component_data.pop("cid", None)
        self.id = component_data.pop("id", None)

        # Name of the component
        self.name = component_data.pop("name", str(self.id))

        # Tag (e.g. BACNET/source name)
        self.tag = component_data.pop("tag", None)

        # Data for the component (PreHEAT_API.Data)
        self.data = None

        # Factor to divide by to obtain a 'standard' unit
        self.std_unit_divisor = component_data.pop("stdUnitDivisor", 1)

        # Standard unit for this type of component
        self.std_unit = component_data.pop("stdUnit", "")

    @property
    def std_unit_devisor(self):
        """DEPRECATED: use std_unit_divisor instead"""
        warn_deprecation_pending("Deprecated: use std_unit_divisor instead")
        return self.std_unit_divisor

    def __repr__(self) -> str:
        """String representation"""
        return f"{type(self).__name__}({self.name}, {self.tag})"



"""
Module defining exceptions used in the package
"""


# For backwards compatibility
class MissingConfigurationFile(Exception):
    pass


class InvalidSchedule(Exception):
    pass


# / backwards compatibility
class NotFoundError(Exception):
    pass


class APIKeyMissingError(Exception):
    pass


class AccessDeniedError(Exception):
    pass


class APIDataExtractError(Exception):
    pass


class MissingConfigurationFileError(MissingConfigurationFile):
    pass


class NoDataError(Exception):
    pass


class NoComfortProfileError(NoDataError):
    pass


class InvalidScheduleError(InvalidSchedule):
    pass



"""
Methods to ensure backwards-compatibility

The contents of this module will be deprecated in future versions
"""

from datetime import datetime

from preheat_open.logging import Logging

from .types import TYPE_DATETIME_INPUT


def warn_deprecation_pending(message: str) -> None:
    """
    Warns that a method/feature will be deprecated
    Note to developers: make sure to use this one as it streamlines the code and makes deprecation easily traceable

    :param message: message to print as warning
    """
    Logging().warning(DeprecationWarning(message))


def load_parameter_old_naming(
    start: TYPE_DATETIME_INPUT,
    end: TYPE_DATETIME_INPUT,
    resolution: str,
    postfix: str = "date",
    **kwargs,
) -> tuple[TYPE_DATETIME_INPUT, TYPE_DATETIME_INPUT, str]:
    """
    Checks for old parameter namings as start_date, end_date and time_resolution and forwards them to their respective
    new namings, i.e. start, end and resolution. Use by passing both new parameters down and a kwargs. Old namings will
    be caught and values passed on while a warning about the deprecation of the old namings is raised.
    """

    d = dict(start=start, end=end, resolution=resolution)
    for param, alt_name in {
        "start": f"start_{postfix}",
        "end": f"end_{postfix}",
        "resolution": "time_resolution",
    }.items():
        if alt_name in kwargs:
            d[param] = kwargs[alt_name]
            warn_deprecation_pending(
                f"Use of parameter '{alt_name}' is deprecated. Use '{param}' instead."
            )
    return d["start"], d["end"], d["resolution"]



import pandas as pd


def test_anonymised_api_usage():
    from preheat_open import set_api_key
    from preheat_open.building import Building, available_buildings
    from preheat_open.test import ANONYMISED_API_KEY, SHORT_TEST_PERIOD

    set_api_key(ANONYMISED_API_KEY)

    df = available_buildings()
    assert isinstance(df, pd.DataFrame)
    assert len(df) == 1, "The API key has access to too many buildings"

    b = Building(int(df["locationId"][0]))
    assert isinstance(b, Building)
    assert b.location["address"] == ""
    assert b.location["organizationName"] == ""
    assert b.location["parentOrganizationName"] == ""

    b.load_data(*SHORT_TEST_PERIOD)



import os
import uuid

from preheat_open.cache import delete_from_cache, load_from_cache, save_to_cache
from preheat_open.logging import Logging


def test_caching():
    try:
        os.chdir("/tmp")
    except Exception as e:
        Logging().warning(
            "Failed to change dir to /tmp (maybe you are on Windows) // Error: "
            + str(e)
        )

    a = "This is a test cache content"
    id = "test_cache"
    path_out = ".preheat_test_cache_" + str(uuid.uuid4())

    delete_from_cache(id, path=path_out)
    __, nok = load_from_cache(id)
    assert nok is False

    save_to_cache(a, id, path=path_out)
    a_cached, ok = load_from_cache(id, path=path_out)
    assert ok is True
    assert a_cached == a

    delete_from_cache(id, path=path_out)
    no_data, nok = load_from_cache(id, path=path_out)
    assert nok is False

    delete_from_cache(id, path=path_out)
    save_to_cache(a, id, path=path_out)
    a_cached, ok = load_from_cache(id, path=path_out)
    assert ok is True
    assert a_cached == a
    delete_from_cache(id, path=path_out)

    os.rmdir(path_out)



import datetime

import numpy as np
import pandas as pd
import pytest
from requests.models import Response

from preheat_open import test
from preheat_open.helpers import sanitise_datetime_input
from preheat_open.setpoints import InvalidScheduleError

IGNORE_UNIT_INACTIVE_WARNING = "ignore:Warning"
ENABLE_SCHEDULE_VALUE_CHECKING = (
    False  # Disabled in order to support parallel runs without conflicts
)


class TestControlUnit(test.PreheatTest):
    def test_query(self, control_unit):
        assert control_unit is not None

    @pytest.mark.filterwarnings(IGNORE_UNIT_INACTIVE_WARNING)
    def test_get_schedule(self, control_unit):
        start_time = "2020-01-01T00:00:00+02:00"
        end_time = "2020-01-01T12:00:00+02:00"
        schedule_data = control_unit.get_schedule(start_time, end_time)
        assert schedule_data is not None
        assert not schedule_data.empty

        start_time = "1970-06-01T00:00:00+02:00"
        end_time = "1970-06-01T12:00:00+02:00"
        schedule_data = control_unit.get_schedule(start_time, end_time)
        assert schedule_data is not None
        assert schedule_data.empty

    @pytest.mark.filterwarnings(IGNORE_UNIT_INACTIVE_WARNING)
    def test_request_schedule(self, control_unit):
        start_time = sanitise_datetime_input("2020-01-01T00:00:00+02:00")
        end_time = sanitise_datetime_input("2020-01-01T00:05:00+02:00")

        with pytest.warns(RuntimeWarning):
            response = self.send_schedule(
                control_unit, start_time, end_time, 1, 0, 1, 0, 1, 0
            )
            assert response.status_code == 200
            schedule = control_unit.get_schedule(
                start_time, end_time + datetime.timedelta(seconds=1)
            )  # type: pd.DataFrame
            if ENABLE_SCHEDULE_VALUE_CHECKING:
                self.check_schedule(schedule, 1, 0, 1, 0, 1, 0)

            response = self.send_schedule(
                control_unit, start_time, end_time, 0, 1, 0, 1, 0, 1
            )
            assert response.status_code == 200
            if ENABLE_SCHEDULE_VALUE_CHECKING:
                schedule = control_unit.get_schedule(
                    start_time, end_time + datetime.timedelta(seconds=1)
                )
                self.check_schedule(schedule, 0, 1, 0, 1, 0, 1)

            response = self.send_schedule_without_mode(
                control_unit, start_time, end_time, 0, 1, 0, 1, 0, 2
            )
            assert response.status_code == 200
            if ENABLE_SCHEDULE_VALUE_CHECKING:
                schedule = control_unit.get_schedule(
                    start_time, end_time + datetime.timedelta(seconds=1)
                )
                self.check_schedule(schedule, 0, 1, 0, 1, 0, 2)

        # Check raises missing values
        with pytest.raises(InvalidScheduleError):
            self.send_schedule(control_unit, start_time, end_time, 0, np.nan)

        # Check raises infinite values
        with pytest.raises(InvalidScheduleError):
            self.send_schedule(control_unit, start_time, end_time, 0, +np.nan)

        # Check raises negative infinite values
        with pytest.raises(InvalidScheduleError):
            self.send_schedule(control_unit, start_time, end_time, 0, -np.inf)

    @pytest.mark.filterwarnings(IGNORE_UNIT_INACTIVE_WARNING)
    def check_schedule(self, schedule, *values):
        assert schedule is not None
        assert schedule.values is not None
        assert len(schedule.values) == len(values)
        for x, y in zip(schedule.values, values):
            assert x[0] == y

    @pytest.mark.filterwarnings(IGNORE_UNIT_INACTIVE_WARNING)
    def send_schedule(self, control_unit, start_time, end_time, *values) -> Response:
        t_range = pd.date_range(start_time, end_time, len(values))
        start_times = [pd.to_datetime(t) for t in t_range]
        schedule = {
            "value": values,
            "startTime": start_times,
            "operation": len(values) * ["NORMAL"],
        }
        new_schedule = pd.DataFrame(schedule)
        new_schedule.set_index("startTime", inplace=True)
        return control_unit.request_schedule(new_schedule)  # type: Response

    @pytest.mark.filterwarnings(IGNORE_UNIT_INACTIVE_WARNING)
    def send_schedule_without_mode(
        self, control_unit, start_time, end_time, *values
    ) -> Response:
        t_range = pd.date_range(start_time, end_time, len(values))
        start_times = [pd.to_datetime(t) for t in t_range]
        schedule = {
            "value": values,
            "startTime": start_times,
        }
        new_schedule = pd.DataFrame(schedule)
        new_schedule.set_index("startTime", inplace=True)
        return control_unit.request_schedule(new_schedule)  # type: Response

    def test_get_zones(self, control_unit):
        from preheat_open.building_unit import BuildingUnit
        from preheat_open.zone import Zone

        z_A1 = control_unit.get_zones()
        pars = control_unit.parents()
        assert len(z_A1) == 1
        assert isinstance(z_A1[0], Zone)
        assert z_A1[0].name == "Zone_A1"
        assert len(pars) == 1
        assert isinstance(pars[0], BuildingUnit)



"""
Module defining units and their auxiliary functions
"""

from datetime import datetime
from enum import Enum
from typing import Optional, Union

import pandas as pd

from .api import TIMEZONE
from .backwards_compatibility import load_parameter_old_naming, warn_deprecation_pending
from .component import Component
from .data import load_box_data, load_model_data
from .exceptions import NotFoundError
from .helpers import (
    list_to_string,
    now,
    sanitise_datetime_input,
    time_resolution_aliases,
    timedelta,
    utc,
)
from .logging import Logging
from .types import TYPE_DATETIME_INPUT


class DataAlreadyPresentInUnitWarning(RuntimeWarning):
    pass


def populate_units(*args, **kwargs):
    """DEPRECATED method"""
    warn_deprecation_pending("This method is deprecated (and not doing anything)")


class Unit(object):
    """Defines a unit in the PreHEAT sense"""

    def __init__(self, unit_type: str, unit_data: dict, load_data_by: str = "id"):
        # Type
        self.unit_type: str = unit_type
        self.unit_subtype: Optional[str] = unit_data.pop("type", None)

        # Identifier of the unit
        self.id = unit_data.pop("id", None)
        # Name of the unit
        self.name: str = unit_data.pop("name", f"{self.unit_type}_{self.id}")
        if not self.name:
            # If the name does not exist, create one out of the ID and the unit type
            self.name = f"{self.unit_type}_{self.id}"

        # list of components in the unit (PreHEAT_API.Component)
        self.components: list[Component] = self.__populate_components(unit_data)

        # Time series cache
        self.data: pd.DataFrame = pd.DataFrame()

        # State cache
        self._state: pd.DataFrame = pd.DataFrame()

        # Choose how to load the data
        self.__loads_data_by = load_data_by

    def __repr__(self) -> str:
        """String representation of the unit"""
        return f"{self.unit_type}({self.name})"

    def __populate_components(self, unit_data: dict) -> list[Component]:
        components = []
        try:
            keys_to_extract = []
            for key, value in unit_data.items():
                # Component properties: a dict w. key 'cid'
                if isinstance(value, dict) and "cid" in value:
                    keys_to_extract.append(key)
            for key in keys_to_extract:
                comp = unit_data.pop(key)
                # Let 'name' be PreHEAT name, and tag be BACNET/source name
                comp["tag"] = comp["name"]
                comp["name"] = key
                # Add PreHEAT name as description
                components.append(Component(comp))

        except TypeError:
            pass

        return components

    def load_data(
        self,
        start: TYPE_DATETIME_INPUT,
        end: TYPE_DATETIME_INPUT,
        resolution: Union[str, Enum] = "minute",
        components: Union[list[str], None] = None,
        **kwargs,
    ) -> None:
        """
        Loads data in the unit

        :param start: start of the period to load data for
        :param end: end of the period to load data for
        :param resolution: resolution to load data with (raw, 5min, minute, hour, day, week, month, year)
        :param components: components to restrict the loading to (None: no restriction)
        :param kwargs: /
        :return: /
        """
        start, end, resolution = load_parameter_old_naming(
            start, end, resolution, **kwargs
        )
        self._warn_if_data_is_loaded()
        if self.__loads_data_by == "id":
            self.data = load_model_data(
                self.get_all_component_ids(components=components),
                start,
                end,
                resolution,
            )
        elif self.__loads_data_by == "cid":
            self.data = load_box_data(
                self.get_all_component_cids(components=components),
                start,
                end,
                resolution,
            )
        else:
            raise ValueError("")
        self._ensure_continuity_of_data(resolution)

    def _ensure_continuity_of_data(self, resolution: str) -> None:
        if self.data.empty or (resolution == "raw"):
            return

        time_alias = time_resolution_aliases(resolution)

        if resolution in ["day", "week", "month", "year"]:
            # Reindexing to local time prior to frequency conversion
            self.data.index = self.data.index.tz_convert(TIMEZONE)
            self.data = self.data.asfreq(freq=time_alias)
            self.data.index = self.data.index.tz_convert(utc)
        else:
            self.data = self.data.asfreq(freq=time_alias)

    def _warn_if_data_is_loaded(self):
        if self.data.empty is False:
            Logging().warning(
                DataAlreadyPresentInUnitWarning(
                    f"Data was already present in unit (id={self.id}, name={self.name}, type={self.unit_type})"
                )
            )

    def clear_data(self, **kwargs) -> None:
        """
        Clears data from the unit

        :param kwargs: /
        :return: /
        """
        self.data = self.data[0:0]

    def cquery(self, name: str) -> Component:
        return [component for component in self.components if component.name == name][0]

    def _select_components(self, components: Union[list, None] = None) -> list:
        if components is None:
            return self.components
        else:
            if isinstance(components, str):
                components = [components]
            return [
                component
                for component in self.components
                if component.name in components
            ]

    def get_all_component_cids(
        self, prefix: bool = False, components: Union[list, None] = None
    ):
        prefix = "{}.".format(self.name) if prefix else ""
        comps = self._select_components(components=components)
        return {str(c.cid): prefix + c.name for c in comps}

    def get_all_component_ids(
        self, prefix: bool = False, components: Union[list, None] = None
    ) -> dict[str, str]:
        prefix = "{}.".format(self.name) if prefix else ""
        comps = self._select_components(components=components)
        return {str(c.id): prefix + c.name for c in comps}

    def get_all_component_details(
        self, prefix: bool = False, components: Union[list, None] = None
    ) -> list[dict[str, Union[str, float, None]]]:
        prefix = f"{self.name}." if prefix else ""
        comps = self._select_components(components=components)

        return [
            {
                "id": str(c.id),
                "cid": str(c.cid),
                "name": prefix + c.name,
                "stdUnitDivisor": c.std_unit_divisor,
                "stdUnitDevisor": c.std_unit_divisor,  # Kept for backwards compatibility (despite typo)
                "stdUnit": c.std_unit,
            }
            for c in comps
        ]

    def get_component(self, name: str) -> Component:
        """
        Returns the component with the given name

        :param name: name of the component to look for
        :raise: NotFoundError if no component with the given name is found
        :return: component with the chosen name
        """
        try:
            return [
                component for component in self.components if component.name == name
            ][0]
        except Exception as e:
            raise NotFoundError(
                f"The component ({name}) does not exist in the unit ({self.name})."
            ) from e

    def has_component(self, component_name: str) -> bool:
        """
        Tells if a unit a given component

        :param component_name: name of the component to look for
        :return: True if the unit has a component with the given name, False otherwise
        """
        try:
            c = self.get_component(component_name)
            out = isinstance(c, Component)
        except NotFoundError as e:
            out = False
        return out

    def has_data(
        self,
        component: Union[str, None] = None,
        check_not_null: bool = True,
    ) -> bool:
        """
        Tells whether the unit has data

        :param component: if non-None, tells whether data is available for the specific component
        :param check_not_null: if True, checks that the data is not null
        :return: True if data is available, False otherwise
        """
        if component is None:
            return len(self.data) > 0
        elif self.has_component(component) is False:
            return False
        elif self.data.empty is True:
            return False
        elif component in self.data.columns and self.data.shape[0] > 0:
            return self.data[component].notnull().any() if check_not_null else True
        else:
            return False

    # Methods to manage unit state
    def load_state(
        self,
        seconds_back: int = 300,
        t_now: Union[datetime, None] = None,
        resolution_overwrite: Optional[str] = None,
    ) -> None:
        """
        Loads the state of the unit at a given point in time

        :param seconds_back: number of seconds back prior to t_now to look for data for
            (if none is found in the interval, then value will be NaN for this component)
        :param t_now: now-time override
        :param resolution_overwrite: if specified, then loads the data using that specific resolution
            (by default, data is loaded in "raw" resolution if t_now is more recent than 7 days ago, else "5min")
        :return: /
        """
        t_now = now() if t_now is None else sanitise_datetime_input(t_now)
        # Check if recent enough to expect raw data
        if resolution_overwrite is not None:
            resolution = resolution_overwrite
        elif t_now > now() - timedelta(days=7):
            resolution = "raw"
        else:
            resolution = "minute"
        self._state = load_model_data(
            self.get_all_component_ids(),
            t_now - timedelta(seconds=seconds_back),
            t_now,
            resolution,
        )

    def clear_state(self, **kwargs) -> None:
        """
        Clears the state data from the unit

        :param kwargs: /
        :return: /
        """
        self._state = pd.DataFrame()

    def get_state(
        self,
        update: bool = False,
        estimate: str = "last",
        seconds_back: int = 300,
        t_now: Union[datetime, None] = None,
        by: str = "id",
        resolution_overwrite: Optional[str] = None,
    ) -> pd.Series:
        """
        Fetches the state for the given unit (to force refresh it, update is required)

        :param update: if True, loads the state data to update it, else just fetches the state already loaded
        :param estimate: method to use for state estimation
            ("last" for last available value, "mean" for average in period, "median" for median value in period)
        :param seconds_back: number of seconds before t_now to look for when loading data
        :param t_now: now-time override
        :param by: whether to load data by "id" or component id "cid"
        :param resolution_overwrite: if specified, then loads the data using that specific resolution
            (by default, data is loaded in "raw" resolution if t_now is more recent than 7 days ago, else "5min")
        :return: unit state in pandas Series format
        """
        if update is True:
            self.load_state(
                seconds_back=seconds_back,
                t_now=t_now,
                resolution_overwrite=resolution_overwrite,
            )

        if (self._state is None) or self._state.empty:
            if by == "id":
                state = pd.Series(
                    index=self.get_all_component_ids().values(), dtype=float
                )
            elif by == "cid":
                state = pd.Series(
                    index=self.get_all_component_cids().values(), dtype=float
                )
            else:
                raise ValueError(
                    f"Bad value for 'by' input ({by}) must be either id or cid"
                )
        elif estimate == "last":
            state = self._state.ffill().iloc[-1, :]
        elif estimate == "mean":
            state = self._state.mean()
        elif estimate == "median":
            state = self._state.median()
        else:
            raise Exception(f"Invalid value for estimate ({estimate})")

        return state

    def describe(
        self,
        display: bool = True,
        prefix: str = "",
        components: bool = True,
        show_subtype: bool = True,
        **kwargs,
    ) -> str:
        """
        Generate a text-based description of a unit

        :param display: if True, prints the output
        :param prefix: string prefix to each output line (use for tabs, etc)
        :param components: if True, outputs components details
        :param show_subtype: if True, shows the sub-type of units
        :param kwargs: /
        :return: text description of units as string
        """
        if show_subtype:
            subtype = f" / sub-type={self.unit_subtype}"
        else:
            subtype = ""
        out = prefix + f"[{self.unit_type}] {self.name} [id={self.id}{subtype}]"
        if components:
            out += ": " + list_to_string([c.name for c in self.components])
        if display:
            print(out)
        return out



"""
Module defining building units and their auxiliary methods
"""
from __future__ import annotations

import re
from enum import Enum
from typing import Optional, Union

from .backwards_compatibility import load_parameter_old_naming
from .helpers import (
    check_no_remaining_fields,
    convenience_result_list_shortener,
    sanitise_unit_type_input,
)
from .logging import Logging
from .types import TYPE_DATETIME_INPUT
from .unit import Unit
from .weather import Weather


def populate_units(unit_type, units_data, building_ref):
    # Loop over units; create and populate Unit instances
    units = [BuildingUnit(unit_type, unit_i, building_ref) for unit_i in units_data]
    return units


def add_unique_units(units, new_units, ids) -> None:
    for item in new_units:
        if item.id not in ids:
            units.append(item)
            ids.append(item.id)


def exclude_shared_units_from_list(x: list[BaseBuildingUnit]) -> list[BaseBuildingUnit]:
    return [u for u in x if not u.is_shared()]


class BaseBuildingUnit(Unit):
    """Building Unit; an extension of Unit to handle zones and parent units"""

    def __init__(
        self,
        unit_type: str,
        unit_data: dict[str, Union[str, float, bool, None]],
        building_ref,
        load_data_by: str = "id",
    ):
        super().__init__(unit_type, unit_data, load_data_by=load_data_by)

        # Building reference
        self.building = building_ref

        # Zones covered (Attempt to first extract 'zoneIds', then ['zoneId']), and couple zones
        self.zone_ids = []
        self.__zones = []
        zone_ids = unit_data.pop("zoneIds", None)
        if zone_ids is None:
            if "zoneId" in unit_data:
                zone_ids = [unit_data.pop("zoneId", None)]
            else:
                zone_ids = []
        [self.__add_zone(zone_id) for zone_id in zone_ids]
        [z_i.add_coupled_unit(self) for z_i in self.__zones]

        # Indication on whether the unit is shared
        self.__shared = ("shared" in unit_data) and unit_data.pop("shared")

    def __add_zone(self, zone_id):
        self.zone_ids.append(zone_id)
        self.__zones += self.building.get_zones([zone_id])

    def parents(self) -> list[BaseBuildingUnit]:
        return [
            self.building.unit_graph.nodes[item[0]]["unit"]
            for item in self.building.unit_graph.pred[self.id].items()
        ]

    @property
    def weather(self) -> Weather:
        return self.building.weather

    # Taken from open API: building_unit.py
    def query_parent_units(
        self,
        unit_type: Optional[str] = None,
        name: Optional[str] = None,
        exclude_shared: bool = False,
        _depth: int = 5,
    ) -> list[BaseBuildingUnit]:
        # if no levels left to traverse no parents to output
        if _depth < 1 or type(_depth) is not int:
            return []
        # If we pass a unit_type, check if we do regex search or strict search
        if unit_type:
            if unit_type[0] == "?":
                r_type = re.compile(unit_type[1:])
                type_match = lambda t: r_type.search(t)
            else:
                type_match = lambda t: t == unit_type

        # If we pass name, check if we do regex search or strict search
        if name:
            # Check if we do strict search or search by regex:
            if name[0] == "?":
                r_name = re.compile(name[1:])
                name_match = lambda n: r_name.search(n)
            else:
                name_match = lambda n: n == name

        try:
            # If we pass unit_type and name
            if unit_type and name:
                result = [
                    unit
                    for unit in self.parents()
                    if name_match(unit.name) and type_match(unit.unit_type)
                ]

            elif name:
                result = [unit for unit in self.parents() if name_match(unit.name)]

            elif unit_type:
                result = [unit for unit in self.parents() if type_match(unit.unit_type)]

        except Exception as e:
            Logging().warning(f"Error when querying parent units ({str(e)})")
            result = []

        # add one step up
        if _depth > 1:
            res_id = [p.id for p in result]
            new_depth = _depth - 1
            for parent in self.parents():
                if type(parent) is BuildingUnit:
                    next = parent.query_parent_units(
                        unit_type=unit_type, name=name, _depth=new_depth
                    )
                    add_unique_units(result, next, res_id)

        if exclude_shared:
            result = exclude_shared_units_from_list(result)

        return result

    def qpu(self, *args, **kwargs) -> Union[BaseBuildingUnit, list[BaseBuildingUnit]]:
        result = self.query_parent_units(*args, **kwargs)
        return convenience_result_list_shortener(result)

    def get_zones(self):
        return self.__zones

    def query_units(self, *args, **kwargs) -> list[BaseBuildingUnit]:
        return []

    def qu(self, *args, **kwargs) -> Union[BaseBuildingUnit, list[BaseBuildingUnit]]:
        result = self.query_units(*args, **kwargs)
        return convenience_result_list_shortener(result)

    # Experimental functions below, use with care
    def is_shared(self) -> bool:
        return self.__shared


class BuildingUnit(BaseBuildingUnit):
    """Building Unit; an extension of BaseBuildingUnit to handle subUnits"""

    def __init__(self, unit_type, unit_data, building_ref):
        from .control_unit import ControlUnit

        super().__init__(unit_type, unit_data, building_ref)

        # Sub-units (preheat_open.BuildingUnit list)
        self.sub_units = []
        self.__sub_unit_ids = []
        unit_types = [
            "subVentilations",
            "secondaries",
            "subHeating",
            "heatingCoils",
            "coolingCoils",
            "heatRecovery",
            "secondaryCirculation",
            "secondaryTank",
            "secondaryHeatExchanger",
            "secondaryHeatExchangerWithTank",
            "subMeters",
            "radiators",
            "floorHeatingLoops",
        ]

        [
            self.add_sub_unit(
                populate_units(unit_type, unit_data.pop(unit_type, []), building_ref)
            )
            for unit_type in unit_types
        ]

        # Controls
        _controls = unit_data.pop("controls", [])
        [self.add_sub_unit(ControlUnit(control, building_ref)) for control in _controls]

        # These units and maters need to be linked at a later stage when available in the building
        self._related_sub_units_refs = []
        self._related_sub_units_refs += unit_data.pop("heatingCoilIds", [])
        self._related_sub_units_refs += unit_data.pop("coolingCoilIds", [])
        # self.sub_unit_refs += unit_data.pop("zoneIds", [unit_data.pop("zoneId", [])])
        self._related_meters_refs = []
        meter_labels = [
            "electricityId",
            "mainId",
            "heatingId",
            "hotWaterId",
            "coldWaterId",
        ]
        for label in meter_labels:
            if label in unit_data and unit_data[label]:
                self._related_meters_refs.append(unit_data.pop(label))

        # For now, this is a protected attribute
        self._meters = []

        check_no_remaining_fields(unit_data, debug_helper="building_unit_data")

    def add_sub_unit(self, u):
        if isinstance(u, list):
            [self.add_sub_unit(u_i) for u_i in u]
        elif isinstance(u, Unit):
            if u.id not in self.__sub_unit_ids:
                self.sub_units += [u]
                self.__sub_unit_ids += [u.id]
        else:
            TypeError(f"Invalid type for input u ({type(u)}")

    def load_data(
        self,
        start: TYPE_DATETIME_INPUT,
        end: TYPE_DATETIME_INPUT,
        resolution: Union[str, Enum] = "minute",
        load_children=False,
        components=None,
        **kwargs,
    ) -> None:
        start, end, resolution = load_parameter_old_naming(
            start, end, resolution, **kwargs
        )
        super().load_data(start, end, resolution=resolution, components=components)
        if load_children is True:
            [
                u_i.load_data(
                    start,
                    end,
                    resolution=resolution,
                    load_children=True,
                )
                for u_i in self.children()
            ]

    def clear_data(self, clear_children: bool = False) -> None:
        super().clear_data()
        if clear_children:
            [u_i.clear_data(clear_children=True) for u_i in self.children()]

    def children(self) -> list[BaseBuildingUnit]:
        return [
            self.building.unit_graph.nodes[item[0]]["unit"]
            for item in self.building.unit_graph.succ[self.id].items()
        ]

    def query_units(
        self,
        unit_type: Optional[str] = None,
        name: Optional[str] = None,
        exclude_shared: bool = False,
        _depth: int = 10,
    ) -> list[BaseBuildingUnit]:
        # If we pass a unit_type, check if we do regex search or strict search
        if unit_type:
            if unit_type[0] == "?":
                r_type = re.compile(unit_type[1:])
                type_match = lambda t: r_type.search(t)
            else:
                unit_type = sanitise_unit_type_input(unit_type)
                type_match = lambda t: t == unit_type

        # If we pass name, check if we do regex search or strict search
        if name:
            # Check if we do strict search or search by regex:
            if name[0] == "?":
                r_name = re.compile(name[1:])
                name_match = lambda n: r_name.search(n)
            else:
                name_match = lambda n: n == name

        try:
            # Now ensuring that all children units are queried through (not only sub-units)
            candidate_units = self.children()
            children_units_id = [u.id for u in candidate_units]
            for u in self.sub_units + self._meters:
                if u.id not in children_units_id:
                    children_units_id += [u.id]
                    candidate_units += [u]

            # If we pass unit_type and name
            if unit_type and name:
                result = [
                    unit
                    for unit in candidate_units
                    if name_match(unit.name) and type_match(unit.unit_type)
                ]

            elif name:
                result = [unit for unit in candidate_units if name_match(unit.name)]

            elif unit_type:
                result = [
                    unit for unit in candidate_units if type_match(unit.unit_type)
                ]

            # Querying also through children units
            for u_i in candidate_units:
                next_depth = _depth - 1
                result += u_i.query_units(
                    unit_type=unit_type, name=name, _depth=next_depth
                )

            if exclude_shared:
                result = exclude_shared_units_from_list(result)

            return result

        except Exception as e:
            Logging().warning(f"Error when querying parent units ({str(e)})")
            return []

    def describe(
        self, display: bool = True, prefix: str = "", children: bool = True, **kwargs
    ) -> str:
        out = super().describe(display=display, prefix=prefix, **kwargs)
        unit_children = self.children()
        if children is False or len(unit_children) == 0:
            pass  # out += "\n"
        else:
            for u in unit_children:
                out += "\n" + u.describe(
                    display=False, prefix=prefix.replace("-", "") + "\t- ", **kwargs
                )
        if display:
            print(out)
        return out



"""
Managing minimalistic caching of data

The functions automatically take care of creating the folder structure to host the cache files, if needed.
"""


import os
import pickle
from datetime import datetime
from typing import Any, Optional

from preheat_open.logging import Logging


def __create_folder_if_needed(path: str) -> None:
    """
    Ensures that a given output folder (path) exists by creating it if it doesn't

    :param path: str
    :return: /
    """
    if os.path.isdir(path):
        return

    current_path = ""
    folders_in_path = path.split("/")
    for p_i in folders_in_path:
        current_path += f"{p_i}/"
        if (len(current_path) > 1) and (os.path.isdir(current_path) is False):
            Logging().info(f"Creating folder: {current_path}")
            os.mkdir(current_path)


def __cache_filename(
    id: str,
    path: str,
) -> str:
    if not path.endswith("/"):
        path = path + "/"
    __create_folder_if_needed(path)
    if id.endswith(".pkl"):
        out = f"{path}{id}"
    else:
        out = f"{path}{id}.pkl"
    return out


def save_to_cache(data: Any, id: str, path: str = ".cache/") -> None:
    """
    Save given data to an on-disk cache

    :param data: data to cache
    :param id: cache identifier
    :param path: path on which to save the data
    :return: /
    """
    filename = __cache_filename(id, path)
    Logging().debug(f"Caching data to {filename}")
    with open(filename, "wb") as file_id:
        try:
            pickle.dump(data, file=file_id)
        except Exception as e:
            Logging().warning(
                RuntimeWarning(f"Failed to cache to {filename} (exception: {e})")
            )


def load_from_cache(
    id: str,
    expiry_s: Optional[int] = None,
    path: str = ".cache/",
) -> tuple[Any, bool]:
    """
    Loads data from the on-disk cache.

    WARNING: this method loads data from a file using pickle, which may execute code at the time of loading.
        Make sure to load only files that you generated yourself or trust in order to avoid execution of malicious code.

    :param id: cache identifier
    :param expiry_s: acceptable lifetime of the cache (in s)
    :param path: path in which to cache the data
    :return: tuple of:
        data loaded from cache (or None if non-existent) and
        indicator of whether data was loaded
    """
    filename = __cache_filename(id, path)
    if os.path.exists(filename) is False:
        data_should_be_loaded = False
    else:
        if expiry_s is None:
            data_should_be_loaded = True
        elif os.path.getsize(filename) <= 0:
            # If the file is empty, skip the data loading
            data_should_be_loaded = False
        else:
            file_age_s = (datetime.now().timestamp()) - (os.path.getmtime(filename))
            if file_age_s <= expiry_s:
                data_should_be_loaded = True
            else:
                data_should_be_loaded = False
                os.remove(filename)

    data = None
    data_was_loaded = False

    if data_should_be_loaded is True:
        with open(filename, "rb") as file_id:
            try:
                Logging().debug(f"Loading data from cache in {filename}")
                data = pickle.load(file=file_id)
                data_was_loaded = True
            except Exception as e:
                Logging().warning(
                    RuntimeWarning(
                        f"Failed to load from cache in {filename} (exception: {e})"
                    )
                )

    return data, data_was_loaded


def delete_from_cache(id: str, path: str = ".cache/") -> None:
    """
    Deletes data from the on-disk cache

    :param id: cache identifier
    :param path: path in which to cache the data
    :return: /
    """

    filename = __cache_filename(id, path)
    if os.path.exists(filename) is True:
        Logging().debug(f"Deleting cache from {filename}")
        os.remove(filename)



import datetime

import pandas as pd

from preheat_open import test
from preheat_open.api import (
    ApiSession,
    api_get,
    api_put,
    api_string_to_datetime,
    datetime_to_api_string,
    ids_list_to_string,
)


class TestApi(test.PreheatTest):
    def test_load_api_key(self):
        api_key = ApiSession().get_api_key()
        assert len(api_key) > len("Apikey ")
        assert api_key == "Apikey " + test.API_KEY

    def test_api_string_to_datetime(self):
        dt = api_string_to_datetime("2021-05-11T08:09:13.370Z")
        assert isinstance(dt, datetime.datetime)
        assert dt.year == 2021
        assert dt.month == 5
        assert dt.day == 11
        assert dt.hour == 8
        assert dt.minute == 9
        assert dt.second == 13
        assert dt.microsecond == 370000
        assert str(dt.tzinfo) == "UTC"

        dt = api_string_to_datetime("2020-04-10T07:08:12.369+02:00")
        assert isinstance(dt, datetime.datetime)
        assert dt.year == 2020
        assert dt.month == 4
        assert dt.day == 10
        assert dt.hour == 7
        assert dt.minute == 8
        assert dt.second == 12
        assert dt.microsecond == 369000
        assert str(dt.tzinfo) == "UTC+02:00"

    def test_datetime_to_api_string(self):
        dt = api_string_to_datetime("2021-05-11T08:09:13.370Z")
        assert datetime_to_api_string(dt) == "2021-05-11T08:09:13.370000+00:00"

        dt = api_string_to_datetime("2020-04-10T07:08:12.369+02:00")
        assert datetime_to_api_string(dt) == "2020-04-10T07:08:12.369000+02:00"

    def test_api_get(self, building_id):
        response = api_get(f"locations/{building_id}")
        assert response is not None
        assert "building" in response

    def test_api_put(self, control_unit_id):
        payload = {
            "schedule": [
                {
                    "startTime": "2020-01-01T00:00:00+02:00",
                    "value": 1,
                    "operation": "NORMAL",
                },
                {
                    "startTime": "2020-01-01T00:01:00+02:00",
                    "value": 0,
                    "operation": "NORMAL",
                },
                {
                    "startTime": "2020-01-01T00:02:00+02:00",
                    "value": 1,
                    "operation": "NORMAL",
                },
                {
                    "startTime": "2020-01-01T00:03:00+02:00",
                    "value": 0,
                    "operation": "NORMAL",
                },
                {
                    "startTime": "2020-01-01T00:04:00+02:00",
                    "value": 1,
                    "operation": "NORMAL",
                },
                {
                    "startTime": "2020-01-01T00:05:00+02:00",
                    "value": 0,
                    "operation": "NORMAL",
                },
            ]
        }
        response = api_put(f"controlunit/{control_unit_id}/setpoint", payload)
        pass

    def test_ids_list_to_string(self):
        assert ids_list_to_string([1, 2, 3], separator=",") == "1,2,3"
        assert ids_list_to_string([1, 2, 3], separator=";") == "1;2;3"
        assert ids_list_to_string(pd.Series([1, 2, 3]), separator=",") == "1,2,3"
        assert (
            ids_list_to_string(pd.Series([1, 2, 3]).to_numpy(), separator=",")
            == "1,2,3"
        )



from preheat_open.helpers import (
    convenience_result_list_shortener,
    sanitise_unit_type_input,
)


def test_sanitisation():
    assert sanitise_unit_type_input("heatPumps") == "heatPumps"
    assert sanitise_unit_type_input("localWeatherStations") == "localWeatherStations"

    # Test those with actual changes
    assert sanitise_unit_type_input("heatPump") == "heatPumps"
    assert sanitise_unit_type_input("localWeatherStation") == "localWeatherStations"
    assert sanitise_unit_type_input("secondary") == "secondaries"


def test_convenience():
    assert convenience_result_list_shortener([]) is None
    assert convenience_result_list_shortener([123]) == 123
    assert convenience_result_list_shortener([123, 456]) == [123, 456]
    assert convenience_result_list_shortener([123, 456, 789]) == [123, 456, 789]



import pytest

from preheat_open import test
from preheat_open.helpers import now, timedelta


class TestPrice(test.PreheatTest):
    def test_building(self, building):
        assert building is not None

        sps = building.get_supply_points()
        assert sps is not None, "Failure in supply points"

        comps = sps[0].get_price_components()
        assert comps is not None, "Failure in price components extraction"

        # Test backwards compatibility
        with pytest.warns(DeprecationWarning):
            sps[0].load_price_data(start_time=now() - timedelta(days=1), end_time=now())
            data, df = sps[0].get_price_data()
            assert data is not None, "Failure in data"
            assert df is not None, "Failure in data labelling"

        sps[0].clear_price_data()
        sps[0].load_price_data(start=now() - timedelta(days=1), end=now())
        data2, df2 = sps[0].get_price_data()
        assert data2 is not None, "Failure in data"
        assert df2 is not None, "Failure in data labelling"



"""
Module to manage the comfort profiles
"""
import pandas as pd

from .api import api_get, datetime_to_api_string
from .exceptions import NoComfortProfileError
from .helpers import sanitise_datetime_input
from .types import TYPE_DATETIME_INPUT


def fetch_comfort_profiles(
    location_id: int,
    start: TYPE_DATETIME_INPUT,
    end: TYPE_DATETIME_INPUT,
) -> dict[int, pd.DataFrame]:
    """
    Get target temperature using API call
    """
    start = sanitise_datetime_input(start)
    end = sanitise_datetime_input(end)
    package = api_get(
        f"locations/{location_id}/comfortprofile/setpoints",
        payload={
            "start_time": datetime_to_api_string(start),
            "end_time": datetime_to_api_string(end),
        },
        out="json",
    )["comfortProfiles"]
    if not package:
        raise NoComfortProfileError(
            "No comfort profile available for location in period"
        )

    data = {}
    for comfort_profile in package:
        zdf = pd.DataFrame(comfort_profile["setpoints"]).set_index("time")
        zdf.index = pd.DatetimeIndex(zdf.index)
        zdf["name"] = comfort_profile["name"]
        comfort_profile_id = comfort_profile["id"]

        # Fixing API issue by cropping the data to the focus period
        data[comfort_profile_id] = zdf[start:end]

    return data



import pandas as pd

import preheat_open.data
from preheat_open import test
from preheat_open.helpers import sanitise_datetime_input

KWARGS_PANDAS_EQUALS = {"check_like": True}

START_DATE = sanitise_datetime_input("2021-05-01 00:00:00+02:00")
END_DATE = sanitise_datetime_input("2021-05-07 00:00:00+02:00")
COMPONENT_MAP = {
    "32618": "supplyT",
    "32621": "returnT",
    "32622": "flow",
    "32619": "volume",
    "32617": "energy",
    "32620": "power",
}
COMPONENT_MAP_BOX = {"1062545": "x", "1062544": "y"}
COMPONENT_MAP_WEATHER = {
    "18": "Temperature",
    "19": "Humidity",
    "20": "WindDirection",
    "21": "WindSpeed",
    "22": "Pressure",
    "23": "LowClouds",
    "24": "MediumClouds",
    "25": "HighClouds",
    "26": "Fog",
    "27": "WindGust",
    "28": "DewPointTemperature",
    "29": "Cloudiness",
    "30": "Precipitation",
    "31": "DirectSunPower",
    "32": "DiffuseSunPower",
    "33": "SunAltitude",
    "34": "SunAzimuth",
    "35": "DirectSunPowerVertical",
}
TIME_RESOLUTION = "hour"


def _test_dataframe_equality(df1: pd.DataFrame, df2: pd.DataFrame):
    cols = df1.columns
    pd.testing.assert_frame_equal(df1, df2[cols], check_like=False)


class TestData(test.PreheatTest):
    def test_json_csv_model(self):
        """
        Test that the new csv format options is equal to the old json
        """
        path = "units/measurements"
        id_key = "ids"

        resp_json = preheat_open.data.perform_requests(
            path,
            id_key,
            list(COMPONENT_MAP.keys()),
            START_DATE,
            END_DATE,
            TIME_RESOLUTION,
            "json",
        )
        data_json = preheat_open.data.extract_data(
            resp_json, "id", COMPONENT_MAP, "json"
        )

        resp_csv = preheat_open.data.perform_requests(
            path,
            id_key,
            list(COMPONENT_MAP.keys()),
            START_DATE,
            END_DATE,
            TIME_RESOLUTION,
            "csv",
        )
        data_csv = preheat_open.data.extract_data(resp_csv, "id", COMPONENT_MAP, "csv")

        assert data_json is not None
        assert data_csv is not None
        assert not data_json.empty
        assert not data_csv.empty
        _test_dataframe_equality(data_json, data_csv)

    def test_json_csv_box(self):
        """
        Test that the new csv format options is equal to the old json
        """
        path = "measurements"
        id_key = "cids"

        resp_json = preheat_open.data.perform_requests(
            path,
            id_key,
            list(COMPONENT_MAP_BOX.keys()),
            START_DATE,
            END_DATE,
            TIME_RESOLUTION,
            "json",
        )
        data_json = preheat_open.data.extract_data(
            resp_json, "cid", COMPONENT_MAP_BOX, "json"
        )

        resp_csv = preheat_open.data.perform_requests(
            path,
            id_key,
            list(COMPONENT_MAP_BOX.keys()),
            START_DATE,
            END_DATE,
            TIME_RESOLUTION,
            "csv",
        )
        data_csv = preheat_open.data.extract_data(
            resp_csv, "cid", COMPONENT_MAP_BOX, "csv"
        )

        assert data_json is not None
        assert data_csv is not None
        assert not data_json.empty
        assert not data_csv.empty
        _test_dataframe_equality(data_json, data_csv)

    def test_json_csv_weather(self, building_id):
        """
        Test that the new csv format options is equal to the old json
        """
        path = f"weather/{building_id}"
        id_key = "type_ids"

        resp_json = preheat_open.data.perform_requests(
            path,
            id_key,
            list(COMPONENT_MAP_WEATHER.keys()),
            START_DATE,
            END_DATE,
            TIME_RESOLUTION,
            "json",
        )
        data_json = preheat_open.data.extract_data(
            resp_json, "type_id", COMPONENT_MAP_WEATHER, "json"
        )

        resp_csv = preheat_open.data.perform_requests(
            path,
            id_key,
            list(COMPONENT_MAP_WEATHER.keys()),
            START_DATE,
            END_DATE,
            TIME_RESOLUTION,
            "csv",
        )
        data_csv = preheat_open.data.extract_data(
            resp_csv, "type_id", COMPONENT_MAP_WEATHER, "csv"
        )

        assert data_json is not None
        assert data_csv is not None
        assert not data_json.empty
        assert not data_csv.empty
        _test_dataframe_equality(data_json, data_csv)

    def test_load_box_data(self):
        df = preheat_open.data.load_box_data(
            COMPONENT_MAP_BOX, START_DATE, END_DATE, TIME_RESOLUTION
        )
        assert df is not None
        assert not df.empty

    def test_load_model_data(self):
        df = preheat_open.data.load_model_data(
            COMPONENT_MAP, START_DATE, END_DATE, TIME_RESOLUTION
        )
        assert df is not None
        assert not df.empty

    def test_load_weather_data(self, building_id):
        df = preheat_open.data.load_weather_data(
            building_id, COMPONENT_MAP_WEATHER, START_DATE, END_DATE, TIME_RESOLUTION
        )
        assert df is not None
        assert not df.empty



"""
Sub-package containing all test routines to be run with Pytest (parameterized with 'conftest.py')
"""
import os
import warnings
from typing import Union

import pytest

import preheat_open
from preheat_open import ENV_VAR_TRACK_MISSING_DATA_FIELD, running_in_test_mode
from preheat_open.building import Building
from preheat_open.unit import Unit

# Setting up a test API key, which is only valid for a dummy test installation
TEST_LOCATION_ID = 2756
API_KEY = "KVkIdWLKac5XFLCs2loKb7GUitkTL4uJXoSyUFIZkVgWuCk8Uj"
ANONYMISED_API_KEY = "3xa0SeGXa4WlkrB68qGR9NoDAzVvGdiG3XAabKu6n7n5qQTDkL"

# Warning the user that this module is not meant to be used for non test-related activities
if not running_in_test_mode():
    warnings.warn(
        """

The module 'preheat_open.test' is not meant to be imported and actively used, 
unless you are specifically carrying out a test.

    """
    )

SHORT_TEST_PERIOD = ("2021-05-01T00:00+02:00", "2021-05-02T00:00+02:00", "hour")
MEDIUM_TEST_PERIOD = ("2021-05-01T00:00+02:00", "2021-05-02T00:00+02:00", "hour")


class PreheatTest:
    @pytest.fixture()
    def bypass_api_key(self):
        preheat_open.api.set_api_key(None)
        yield None
        preheat_open.api.set_api_key(API_KEY)


def mark_pytest_package_test():
    preheat_open.helpers.DISPLAY_DEVELOPER_WARNING.cache_clear()
    os.environ[ENV_VAR_TRACK_MISSING_DATA_FIELD] = "true"


def __clearing_forbidden(*args, **kwargs):
    raise RuntimeError("Clearing of data in fixture for data is forbidden")


def forbid_clear_data_on_object(x: Union[Building, Unit]):
    if isinstance(x, Unit):
        x.clear_data = __clearing_forbidden
    elif isinstance(x, Building):
        x.clear_data = __clearing_forbidden
        [forbid_clear_data_on_object(u) for u in x.list_all_units()]
    else:
        raise TypeError("Input must be a Building or a Unit")



"""
This module defines the basics of the interaction with Neogrid's web API
"""
import json
import os
from datetime import datetime
from io import StringIO
from typing import Any, Optional, Union

import numpy as np
import pandas as pd
from dateutil.tz import gettz
from requests import Session
from requests.adapters import HTTPAdapter
from requests.models import Response
from urllib3.util.retry import Retry

import preheat_open

from .backwards_compatibility import warn_deprecation_pending
from .exceptions import (  # For backwards compatibility
    AccessDeniedError,
    APIDataExtractError,
    APIKeyMissingError,
    MissingConfigurationFile,
    MissingConfigurationFileError,
)
from .logging import Logging, write_log_to_profiling_file
from .singleton import Singleton

# Settings of the API
BASE_URL: str = "https://api.neogrid.dk/public/api/v1"
MAX_CIDS_PER_REQ: int = 3
MAX_POINTS_PER_REQ: int = 90000
MAX_IDS_AND_CIDS_PER_REQUEST: int = 100
GET_TIMEOUT_SECONDS: int = 60
PUT_TIMEOUT_SECONDS: int = 10

# Perhaps we can steal this from the OS?
TIMEZONE = gettz("Europe/Copenhagen")
DATETIME_FORMAT: str = "%Y-%m-%dT%H:%M:%S.%fZ"

# Addresses of the configuration file (if changing this, make sure to update the documentation in the README file
USER_LEVEL_CONFIGURATION_FILE: str = os.path.expanduser("~/.preheat/config.json")
MACHINE_LEVEL_CONFIGURATION_FILE: str = os.path.expanduser("/etc/preheat/config.json")
MACHINE_LEVEL_CONFIGURATION_FILE_OPT: str = os.path.expanduser(
    "/etc/opt/preheat/config.json"
)


def configuration_file_path() -> str:
    """
    Determines the path of the configuration file to be used.

    :raise: preheat_open.exceptions.MissingConfigurationFileError if no cofiguration file is available
    :return: full path of the configuration file
    """
    for i in [
        USER_LEVEL_CONFIGURATION_FILE,
        MACHINE_LEVEL_CONFIGURATION_FILE_OPT,
        MACHINE_LEVEL_CONFIGURATION_FILE,
    ]:
        if os.path.exists(i):
            return i
    raise MissingConfigurationFileError()


def api_string_to_datetime(t: str) -> datetime:
    """
    Converts a datetime string from the API to a python datetime

    :param t: datetime in string format
    :return: datetime corresponding to the input
    """
    return datetime.strptime(t, "%Y-%m-%dT%H:%M:%S.%f%z")


def datetime_to_api_string(t: datetime) -> str:
    """
    Converts a python datetime to a string ready for use in the API I/O

    :param t: datetime to convert to string
    :return: a string with a format compatible with the API
    """
    return t.isoformat()


def ids_list_to_string(list2use: list, separator: str = ",") -> str:
    """
    Helper function to turn list into string, e.g. comma separated (default).

    :param list2use: list to convert
    :param separator: separation character (default: ',')
    :return: list in string format
    """

    if isinstance(list2use, list):
        res = separator.join(map(str, list2use))
    elif isinstance(list2use, pd.Series):
        res = separator.join(map(str, list2use.to_list()))
    elif isinstance(list2use, np.ndarray):
        res = separator.join(map(str, list2use.tolist()))
    else:
        res = str(list2use)

    return res


def load_configuration() -> Union[None, dict]:
    """
    Loads the API configuration

    :return: None if no configuration file found, else a dictionary with the keys/values in the configuration file
    """
    config = None
    try:
        # Prefer config file in user home dir, else use the global config file
        config_file = configuration_file_path()
        if os.path.exists(config_file):
            Logging().info(f"Loading config from {config_file}")
            with open(config_file, "r") as fp:
                config = json.load(fp)
    except Exception:
        config = None

    return config


class ApiSession(metaclass=Singleton):
    """Singleton class to handle API connection sessions"""

    def __init__(self):
        # Adding protection against remote closing connection
        retry_strategy = Retry(
            total=3,
            status_forcelist=[429, 500, 502, 503, 504],
            allowed_methods=["GET", "PUT", "POST"],
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        http_session = Session()
        http_session.mount("https://", adapter)
        http_session.mount("http://", adapter)

        self.base_url = BASE_URL
        self.session = http_session
        Logging().debug(f"Creating API session [{self.addr()}]...")

        self.set_api_key()

    def addr(self) -> str:
        """
        Making a representation of the session object for use in logging

        :return: string representing the session
        """
        return hex(id(self.session))

    def get_api_key(self) -> str:
        """
        Getter for the API key of the user

        :return: the API key for the user
        """
        return self.session.headers.get("Authorization")

    def set_api_key(self, api_key: Optional[str] = None) -> None:
        """
        Sets the API key

        :param api_key: value of the API key (if None, loads from the configuration file)
        :return: /
        """
        if api_key is None:
            api_key = self._load_api_key(
                self.get_api_key().split(" ")[1]
                if self.get_api_key() is not None
                else None
            )
        else:
            Logging().debug(f"API key set manually [{self.addr()}]")
        headers = {"Authorization": "Apikey " + str(api_key)}
        self.session.headers.update(headers)

    def set_api_url(self, api_url=None):
        """
        DEPRECATED (only kept for backwards-compatibility)
        """
        warn_deprecation_pending(f"Manual update of the API url is deprecated")

    def __log_api_io(self, msg: str) -> None:
        Logging().debug(msg)
        write_log_to_profiling_file("API-I/O: " + msg)

    def api_get(
        self,
        endpoint: str,
        out: str = "json",
        payload: Optional[dict[str, Any]] = None,
        headers: Optional[dict[str, Any]] = None,
        bare_url: bool = False,
    ):
        """
        Method to send a GET request to the API

        :param endpoint: URL of the endpoint
        :param out: output type (json or csv)
        :param payload: payload for the request
        :param headers: headers to use for the request
        :param bare_url: if True, uses the endpoint as full URL (and skips base URL additions)
        :return: API response in request.Response format
        """
        payload = payload if payload is not None else {}
        headers = headers if headers is not None else {}
        self.__log_api_io(f"GET /{endpoint} [{self.addr()}]")

        if bare_url:
            path = endpoint
        else:
            path = self.base_url + "/" + endpoint

        if out == "csv":
            headers["Accept"] = "text/csv"
        resp = self.session.get(
            path, params=payload, headers=headers, timeout=GET_TIMEOUT_SECONDS
        )

        Logging().debug(f"{resp.status_code} {resp.reason}")
        try:
            resp.raise_for_status()
        except Exception as e:
            msg = """GET - FAILED
            URL: {}
            Payload: {}
            """.format(
                path, str(payload)
            )
            Logging().error(msg, exception=e)
            raise e

        try:
            if out == "json":
                return resp.json()
            elif out == "csv":
                data = StringIO(resp.text)
                return pd.read_csv(data, delimiter=";")
            else:
                raise ValueError("Invalid value for out parameter ({})".format(out))
        except ValueError as e:
            raise e
        except Exception as e:
            raise APIDataExtractError() from e

    def api_put(
        self,
        endpoint: str,
        out: str = "json",
        json_payload: Optional[dict[str, Any]] = None,
        headers: Optional[dict[str, Any]] = None,
        bare_url: bool = False,
    ) -> Response:
        """
        Method to send a PUT request to the API

        :param endpoint: URL of the endpoint
        :param out: output type (json or csv)
        :param json_payload: payload for the request
        :param headers: headers to use for the request
        :param bare_url: if True, uses the endpoint as full URL (and skips base URL additions)
        :return: API response in request.Response format
        """
        json_payload = json_payload if json_payload is not None else {}
        self.__log_api_io(f"PUT /{endpoint} [{self.addr()}]")

        if bare_url:
            path = endpoint
        else:
            path = self.base_url + "/" + endpoint
        r = self.session.put(path, json=json_payload, timeout=PUT_TIMEOUT_SECONDS)
        try:
            r.raise_for_status()
        except Exception as e:
            msg = """PUT - FAILED
            URL: {}
            Payload: {}
            """.format(
                path, str(json_payload)
            )
            Logging().error(msg, exception=e)
            raise e
        return r

    def api_post(
        self,
        endpoint: str,
        out: str = "json",
        json_payload: Optional[dict[str, Any]] = None,
        bare_url: bool = False,
    ) -> Response:
        """
        Method to send a POST request to the API

        :param endpoint: URL of the endpoint
        :param out: output type (json or csv)
        :param json_payload: payload for the request
        :param bare_url: if True, uses the endpoint as full URL (and skips base URL additions)
        :return: API response in request.Response format
        """
        json_payload = json_payload if json_payload is not None else {}
        self.__log_api_io(f"POST /{endpoint} [{self.addr()}]")

        if bare_url:
            path = endpoint
        else:
            path = self.base_url + "/" + endpoint
        r = self.session.post(path, json=json_payload, timeout=PUT_TIMEOUT_SECONDS)
        try:
            r.raise_for_status()
        except Exception as e:
            msg = """POST - FAILED
            URL: {}
            Payload: {}
            """.format(
                path, str(json_payload)
            )
            Logging().error(msg, exception=e)
            raise e
        return r

    def _load_api_key(self, existing: Optional[str]) -> str:
        """
        Load API key from environment variable
        - if API key not present in environment variable it will be set None
        - will then break when trying to access data on a building

        :param existing: the existing API key (None if not set)
        :return: the value of the API key
        """

        # First respect any override by set_api_key
        if os.environ.get("PREHEAT_API_KEY_OVERRIDE") is not None:
            if (
                existing is None
                or os.environ.get("PREHEAT_API_KEY_OVERRIDE") != existing
            ):
                Logging().info(f"Using API key from set_api_key() [{self.addr()}]")
            return os.environ.get("PREHEAT_API_KEY_OVERRIDE")

        # Else, try PREHEAT_API_KEY environment variable
        elif os.environ.get("PREHEAT_API_KEY") is not None:
            if existing is None or os.environ.get("PREHEAT_API_KEY") != existing:
                Logging().info(
                    f"Using API key from environment variable [{self.addr()}]"
                )
            return os.environ.get("PREHEAT_API_KEY")

        # Else, try loading configuration file
        else:
            # If config loaded, set API key accordingly to file
            if preheat_open.config is not None:
                config_keys = preheat_open.config.keys()
                if "API_KEY" in config_keys:
                    if existing is None or preheat_open.config["API_KEY"] != existing:
                        Logging().info(
                            f"Using API key from config file [{self.addr()}]"
                        )
                    return preheat_open.config["API_KEY"]
                elif "PREHEAT_API_KEY" in config_keys:
                    if (
                        existing is None
                        or preheat_open.config["PREHEAT_API_KEY"] != existing
                    ):
                        Logging().info(
                            f"Using API key from config file [{self.addr()}]"
                        )
                    return preheat_open.config["PREHEAT_API_KEY"]
                else:
                    raise Exception(
                        f"""Your configuration file is missing a 'PREHEAT_API_KEY' field"""
                    )

            # No environment variable, no config file, nothing we can do
            else:
                return str(None)

    def _load_api_url(self) -> str:
        """
        Loads the API url

        :return: the base URL of the API
        """
        if os.environ.get("PREHEAT_API_URL_OVERRIDE") is not None:
            url = os.environ.get("PREHEAT_API_URL_OVERRIDE")
            Logging().info(f"Using API url: {url} [{self.addr()}]")
            return url
        else:
            Logging().info(f"Using default API url [{self.addr()}]")
            return BASE_URL


def set_api_key(api_key: Optional[str]) -> None:
    """
    Sets the API key

    :param api_key: value of the API key
    :return: /
    """
    if api_key is None:
        if "PREHEAT_API_KEY_OVERRIDE" in os.environ:
            os.environ.pop("PREHEAT_API_KEY_OVERRIDE")
            Logging().info("API key override unset")
    else:
        os.environ["PREHEAT_API_KEY_OVERRIDE"] = api_key
        Logging().info("API key override set")
    ApiSession().set_api_key()


def api_get(
    endpoint: str,
    out: str = "json",
    payload: Optional[dict[str, Any]] = None,
    headers: Optional[dict[str, Any]] = None,
    bare_url: bool = False,
) -> Response:
    """
    Method to send a GET request to the API

    :param endpoint: URL of the endpoint
    :param out: output type (json or csv)
    :param payload: payload for the request
    :param headers: headers to use for the request
    :param bare_url: if True, uses the endpoint as full URL (and skips base URL additions)
    :return: API response in request.Response format
    """
    return ApiSession().api_get(
        endpoint,
        out=out,
        payload=payload,
        headers=headers,
        bare_url=bare_url,
    )


def api_post(
    endpoint: str,
    out: str = "json",
    json_payload: Optional[dict[str, Any]] = None,
    bare_url: bool = False,
) -> Response:
    """
    Method to send a POST request to the API

    :param endpoint: URL of the endpoint
    :param out: output type (json or csv)
    :param json_payload: payload for the request
    :param bare_url: if True, uses the endpoint as full URL (and skips base URL addition)
    :return: API response in request.Response format
    """
    return ApiSession().api_post(
        endpoint,
        out=out,
        json_payload=json_payload,
        bare_url=bare_url,
    )


def api_put(
    endpoint: str, json_payload: Optional[dict[str, Any]] = None, bare_url: bool = False
) -> Response:
    """
    Method to send a PUT request to the API

    :param endpoint: URL of the endpoint
    :param json_payload: payload for the request
    :param bare_url: if True, uses the endpoint as full URL (and skips base URL addition)
    :return: API response in request.Response format
    """
    return ApiSession().api_put(
        endpoint,
        out="json",
        json_payload=json_payload,
        bare_url=bare_url,
    )


# For backwards compatibility (deprecated methods and classes)
def set_api_url(api_url: str) -> None:
    """
    Only kept for backwards compatibility
    """
    warn_deprecation_pending(f"Manual update of the API url is deprecated")
    ApiSession().set_api_url(api_url)


def __enforce_imports():
    f = [APIKeyMissingError(), AccessDeniedError(), MissingConfigurationFile()]



from __future__ import annotations

from typing import Optional

import pandas as pd

from .building_unit import BaseBuildingUnit
from .exceptions import NoComfortProfileError
from .helpers import check_no_remaining_fields
from .types import TYPE_DATETIME_INPUT


def populate_zones(zones_data, parent_zone=None, building=None):
    # Loop over zones; create and populate Zone instances
    zones = [
        Zone(zone_i, parent_zone=parent_zone, building=building)
        for zone_i in zones_data
    ]
    return zones


class Zone(object):
    """Defines a building zone in the PreHEAT sense"""

    def __init__(self, zone_data, parent_zone=None, building=None):
        self.building = building

        # Identifier of the zone
        self.id = zone_data.pop("id")
        # Name of the zone
        self.name = zone_data.pop("name", str(self.id))
        self.__external_identifier = zone_data.pop("externalIdentifier", "")

        # Floor area of the zone
        self.area = zone_data.pop("zoneArea", None)
        # Whether zone has an external wall (boolean)
        self.has_external_wall = zone_data.pop("hasExternalWall", None)
        # Sub-zones of the zone (list of PreHEAT_API.Zones)
        self.sub_zones = populate_zones(
            zone_data.pop("subZones", []), parent_zone=self, building=self.building
        )
        # Adjacency
        self.adjacent_zones = [
            z["zoneId"] for z in zone_data.pop("adjacentZones", None)
        ]
        # Type of the zone
        self._type = zone_data.pop("type")

        # Parent zone
        self.__parent = parent_zone

        # Adding coupled units
        self.__coupled_units = dict()

        # Adding comfort profile
        self.__comfort_profile_id = zone_data.pop("comfortProfileId", None)

        # Add ventilation context
        self.__ventilation_settings = {
            "supply": zone_data.pop("ventilationSupply", False),
            "exhaust": zone_data.pop("ventilationExhaust", False),
        }

        check_no_remaining_fields(zone_data, debug_helper="zone_data")

    def load_comfort_profile_data(
        self, start: TYPE_DATETIME_INPUT, end: TYPE_DATETIME_INPUT
    ) -> None:
        """
        Loads the comfort profile data in the building object

        :param start: start of the period
        :param end: end of the period
        """
        self.building.load_comfort_profile_data(start, end)

    def clear_comfort_profile_data(self) -> None:
        """
        Clears the comfort profile data loaded in the building object (therefore also the zone)
        """
        self.building.clear_comfort_profile_data()

    @property
    def comfort_profile_data(self) -> pd.DataFrame:
        """
        Accessor to the comfort profile data of the given zone (or its parents).
        This requires prior loading via the *load_comfort_profile_data* method.

        If no comfort profile is defined in the zone and the building only has one,
        then it falls back to the comfort profile at building level.

        Raises NoComfortProfileError if no data is loaded or available

        :return: comfort profile in format: dataframe[setpoint, min, max, value]
        """
        if self.__comfort_profile_id is None:
            if self.__parent is None:
                try:
                    # By default, return comfort profile for building if there is only one
                    out = self.building.comfort_profile_data
                except ValueError as e:
                    raise NoComfortProfileError(
                        "No comfort profile available for zone, and no parent available"
                    ) from e
            else:
                out = self.__parent.comfort_profile_data
        else:
            out = self.building.get_comfort_profile_data(self.__comfort_profile_id)

        return out

    def get_sub_zones(self, zone_ids=None) -> list[Zone]:
        if zone_ids is None:
            res = self.sub_zones
        else:
            res = []
            for z_i in self.sub_zones:
                if z_i.id in zone_ids:
                    res.append(z_i)
                res += z_i.get_sub_zones(zone_ids=zone_ids)
        return res

    def get_parent_zone(self) -> Zone:
        return self.__parent

    def get_type(self) -> tuple[Optional[str], str]:
        """
        Returns the type of the zone ('room', 'stairway', 'corridor', 'bathroom', 'kitchen')
        and its 'dryness' ('dry', 'wet', '?')

        :return: zone_type, wet_or_dry
        """
        if self._type is None:
            return None, "?"
        zone_type = self._type.lower()
        if zone_type in ["room", "stairway", "corridor"]:
            wet_or_dry = "dry"
        elif zone_type in ["bathroom", "kitchen"]:
            wet_or_dry = "wet"
        else:
            wet_or_dry = "?"
        return zone_type, wet_or_dry

    def __repr__(self):
        return "{0}({1})".format(type(self).__name__, self.name)

    def add_coupled_unit(self, unit) -> None:
        unit_type = unit.unit_type
        if unit_type in self.__coupled_units.keys():
            if unit.id in [u_i.id for u_i in self.__coupled_units[unit_type]]:
                pass  # Skip addition of already existing units
            else:
                self.__coupled_units[unit_type].append(unit)
        else:
            self.__coupled_units[unit_type] = [unit]

    def get_units(self, unit_type, sub_zones=True) -> list[BaseBuildingUnit]:
        """
        Returns the coupled units of a given type

        :param unit_type: str
        :param sub_zones: bool (True)
        :return: list(BuildingUnit) or None
        """
        us = self.__coupled_units.get(unit_type)
        if us is None:
            us = []
        if sub_zones is True:
            for z_i in self.sub_zones:
                us_i = z_i.get_units(unit_type, sub_zones=True)
                if len(us_i) > 0:
                    us += us_i
        return us

    def get_unit_types(self, sub_zones=True) -> list[str]:
        """
        Return as list of the coupled unit types
        """
        us = list(self.__coupled_units.keys())
        if sub_zones is True:
            for z_i in self.sub_zones:
                us_i = z_i.get_unit_types(sub_zones=True)
                if us_i is not None:
                    if us is None:
                        us = us_i
                    else:
                        us += us_i
        out = []
        [out.append(k) for k in us if k not in out]
        return out

    def describe(
        self, children: bool = True, prefix: str = "", display: bool = True
    ) -> str:
        z_type, __ = self.get_type()
        type_prefix = "[{}] ".format(z_type) if z_type is not None else ""
        out = prefix + type_prefix + self.name + " [id={}]".format(self.id)
        if children:
            for z in self.sub_zones:
                out += "\n" + z.describe(
                    prefix=prefix.replace("-", "") + "\t- ", display=False
                )
        if display:
            print(out)
        return out

    def query_zones(
        self, zone_id: Optional[int] = None, zone_type: Optional[str] = None
    ) -> list[Zone]:
        out = []
        for z in self.sub_zones:
            if zone_matches(z, zone_id=zone_id, zone_type=zone_type):
                out += [z]
            out += z.query_zones(zone_id=zone_id, zone_type=zone_type)
        return out


def zone_matches(
    z: Zone, zone_id: Optional[int] = None, zone_type: Optional[str] = None
) -> bool:
    by_zone_id = zone_id is not None
    by_zone_type = zone_type is not None
    z_zone_type, __ = z.get_type()

    if by_zone_id and by_zone_type:
        out = (zone_id == z.id) and (zone_type == z_zone_type)
    elif by_zone_id:
        out = zone_id == z.id
    elif by_zone_type:
        out = zone_type == z_zone_type
    else:
        out = False
    return out



import setuptools

import versioneer

package_requirements_file = "requirements.txt"
docs_requirements_file = "docs/requirements.txt"
documentation_requirements = open(docs_requirements_file).read().split("\n")

setuptools.setup(
    name="preheat_open",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="Neogrid and contributors",
    author_email="analytics@neogrid.dk",
    description="Python wrapper for Neogrid Technologies' REST API",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://gitlab.com/neogrid-technologies-public/preheat-open-python",
    project_urls={
        "Bug Tracker": "https://gitlab.com/neogrid-technologies-public/preheat-open-python/-/issues",
        "Documentation": "https://preheat-open.readthedocs.io/en/latest/",
        "Source Code": "https://gitlab.com/neogrid-technologies-public/preheat-open-python",
        "Changelog": "https://gitlab.com/neogrid-technologies-public/preheat-open-python/-/blob/master/RELEASE_NOTES.md",
    },
    packages=setuptools.find_packages(),
    data_files=[("requirements", [package_requirements_file, docs_requirements_file])],
    python_requires=">=3.9",
    install_requires=open(package_requirements_file).read().split("\n"),
    extras_require={
        "doc": documentation_requirements,
        "dev": [
            "setuptools>=42",
            "wheel",
            "pytest",
            "pytest-cov",
            "pytest-xdist",
        ]
        + documentation_requirements,
    },
)
